# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import re
import typing as ta
import urllib.parse

from ..lite.cached import cached_nullary


##


@cached_nullary
def _url_split_host_pat() -> re.Pattern:
    return re.compile('//([^/#?]*)(.*)', re.DOTALL)


def url_split_host(url: str) -> ta.Tuple[ta.Optional[str], str]:
    """splithost('//host[:port]/path') --> 'host[:port]', '/path'."""

    # https://github.com/python/cpython/blob/364ae607d8035db8ba92486ebebd8225446c1a90/Lib/urllib/parse.py#L1143
    if not (m := _url_split_host_pat().match(url)):
        return None, url

    host_port, path = m.groups()
    if path and path[0] != '/':
        path = '/' + path
    return host_port, path


##


def unparse_url_request_path(url: ta.Union[str, urllib.parse.ParseResult]) -> str:
    if isinstance(url, urllib.parse.ParseResult):
        ups = url
    else:
        ups = urllib.parse.urlparse(url)

    return urllib.parse.urlunparse((
        '',
        '',
        ups.path,
        ups.params,
        ups.query,
        ups.fragment,
    ))


def parsed_url_replace(
        url: urllib.parse.ParseResult,
        *,
        scheme: ta.Optional[str] = None,
        netloc: ta.Optional[str] = None,
        path: ta.Optional[str] = None,
        params: ta.Optional[str] = None,
        query: ta.Optional[str] = None,
        fragment: ta.Optional[str] = None,
) -> urllib.parse.ParseResult:
    return urllib.parse.ParseResult(
        scheme if scheme is not None else url.scheme,
        netloc if netloc is not None else url.netloc,
        path if path is not None else url.path,
        params if params is not None else url.params,
        query if query is not None else url.query,
        fragment if fragment is not None else url.fragment,
    )



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import collections  # noqa
import errno
import socket
import typing as ta

from ....io.pipelines.core import IoPipelineMessages
from ....io.pipelines.drivers.sync import SyncSocketIoPipelineDriver
from ....lite.check import check
from ...clients.base import HttpClientContext
from ...clients.base import HttpClientRequest
from ...clients.sync import HttpClient
from ...clients.sync import StreamHttpClientResponse
from ....io.readers import BytesReader
from ....io.readers import BytesReaders
from ...pipelines.clients.clients import IoPipelineHttpClientMessages
from ...pipelines.responses import FullIoPipelineHttpResponse
from .base import BaseIoPipelineHttpClient


##


class IoPipelineHttpClient(HttpClient, BaseIoPipelineHttpClient):
    class _DriverResponseReader:
        def __init__(
                self,
                drv: SyncSocketIoPipelineDriver,
                sock: 'socket.socket',
        ) -> None:
            super().__init__()

            self._drv = drv
            self._sock = sock

            self._done = False

        def read1(self, n: int = -1, /) -> bytes:
            raise NotImplementedError

        def read(self, n: int = -1, /) -> bytes:
            raise NotImplementedError

    #

    def _try_set_nodelay(self, sock: 'socket.socket') -> None:
        try:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        except OSError as e:
            if e.errno != errno.ENOPROTOOPT:
                raise

    def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> StreamHttpClientResponse:
        prepared = self._prepare_request(req)

        sock = socket.create_connection((prepared.parsed_url.host, prepared.parsed_url.port))

        try:
            self._try_set_nodelay(sock)

            drv = SyncSocketIoPipelineDriver(prepared.pipeline_spec, sock)

            drv.enqueue(IoPipelineHttpClientMessages.Request(
                prepared.full_request,
                # aggregate=...
            ))

            response: ta.Optional[FullIoPipelineHttpResponse] = None

            while True:
                if (out := drv.next()) is not None:
                    if isinstance(out, IoPipelineHttpClientMessages.Output):
                        msg = out.msg

                        if isinstance(msg, FullIoPipelineHttpResponse):
                            check.none(response)
                            response = msg

                            drv.enqueue(IoPipelineHttpClientMessages.Close())

                        elif isinstance(msg, (IoPipelineMessages.FinalInput, IoPipelineHttpClientMessages.Close)):
                            pass

                        else:
                            raise TypeError(out)  # noqa

                    else:
                        raise TypeError(out)  # noqa

                if not drv.pipeline.is_ready:
                    break

            #

            response = check.not_none(response)

            response_reader: BytesReader

            if isinstance(response, FullIoPipelineHttpResponse):
                response_reader = BytesReaders.of_bytes(response.body)

                drv.close()
                sock.close()

                def close() -> None:
                    pass

            else:
                response_reader = self._DriverResponseReader(drv, sock)  # type: ignore[unreachable]

                def close() -> None:
                    try:
                        drv.close()
                    finally:
                        sock.close()

            #

            head = check.not_none(response).head

            return StreamHttpClientResponse(
                status=head.status,
                headers=head.headers,
                request=req,
                underlying=drv,
                _stream=response_reader,
                _closer=close,
            )

        except BaseException:
            sock.close()

            raise



# ruff: noqa: UP006 UP007 UP045 UP046
# @omlish-lite
import abc
import typing as ta

from ..lite.abstract import Abstract
from .contexts import CaptureLoggingContext
from .contexts import CaptureLoggingContextImpl
from .contexts import LoggingExcInfoArg
from .levels import LogLevel
from .levels import NamedLogLevel
from .metrics.base import AnyLoggerMetricCollector
from .metrics.base import AnyNopLoggerMetricCollector
from .metrics.base import AsyncLoggerMetricCollector
from .metrics.base import AsyncNopLoggerMetricCollector
from .metrics.base import LoggerMetric
from .metrics.base import LoggerMetricCollector
from .metrics.base import NopLoggerMetricCollector


T = ta.TypeVar('T')


LoggingMsgFn = ta.Callable[[], ta.Union[str, tuple]]  # ta.TypeAlias


##


class AnyLogger(AnyLoggerMetricCollector[T], Abstract, ta.Generic[T]):
    def is_enabled_for(self, level: LogLevel) -> bool:
        return level >= self.get_effective_level()

    @abc.abstractmethod
    def get_effective_level(self) -> LogLevel:
        raise NotImplementedError

    #

    @ta.final
    def isEnabledFor(self, level: LogLevel) -> bool:  # noqa
        return self.is_enabled_for(level)

    @ta.final
    def getEffectiveLevel(self) -> LogLevel:  # noqa
        return self.get_effective_level()

    ##

    # This will be 1 for [Sync]Logger and 0 for AsyncLogger - in sync loggers these methods remain present on the stack,
    # in async loggers they return a coroutine to be awaited and thus aren't actually present when said coroutine is
    # awaited.
    _level_proxy_method_stack_offset: int

    @ta.overload
    def log(self, level: LogLevel, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def log(self, level: LogLevel, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def log(self, level: LogLevel, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def log(self, level: LogLevel, *args, **kwargs):
        return self._log(
            CaptureLoggingContextImpl(
                level,
                stack_offset=self._level_proxy_method_stack_offset,
            ),
            *args,
            **kwargs,
        )

    #

    @ta.overload
    def debug(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def debug(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def debug(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def debug(self, *args, **kwargs):
        return self._log(
            CaptureLoggingContextImpl(
                NamedLogLevel.DEBUG,
                stack_offset=self._level_proxy_method_stack_offset,
            ),
            *args,
            **kwargs,
        )

    #

    @ta.overload
    def info(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def info(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def info(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def info(self, *args, **kwargs):
        return self._log(
            CaptureLoggingContextImpl(
                NamedLogLevel.INFO,
                stack_offset=self._level_proxy_method_stack_offset,
            ),
            *args,
            **kwargs,
        )

    #

    @ta.overload
    def warning(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def warning(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def warning(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def warning(self, *args, **kwargs):
        return self._log(
            CaptureLoggingContextImpl(
                NamedLogLevel.WARNING,
                stack_offset=self._level_proxy_method_stack_offset,
            ),
            *args,
            **kwargs,
        )

    #

    @ta.overload
    def error(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def error(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def error(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def error(self, *args, **kwargs):
        return self._log(
            CaptureLoggingContextImpl(
                NamedLogLevel.ERROR,
                stack_offset=self._level_proxy_method_stack_offset,
            ),
            *args,
            **kwargs,
        )

    #

    @ta.overload
    def exception(self, exc: BaseException, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def exception(self, *, exc_info: LoggingExcInfoArg = True, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def exception(self, msg: str, *args: ta.Any, exc_info: LoggingExcInfoArg = True, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def exception(self, msg: ta.Tuple[ta.Any, ...], *, exc_info: LoggingExcInfoArg = True, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def exception(self, msg_fn: LoggingMsgFn, *, exc_info: LoggingExcInfoArg = True, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def exception(self, *args, exc_info=True, **kwargs):
        if not args:
            if not exc_info:
                raise TypeError('exc_info=False is not allowed when no args are passed')
            args = ((),)
        elif len(args) == 1:
            if isinstance(arg0 := args[0], BaseException):
                if exc_info is not True:  # noqa
                    raise TypeError(f'exc_info={exc_info!r} is not allowed when exc={arg0!r} is passed')
            args, exc_info = ((),), arg0

        return self._log(
            CaptureLoggingContextImpl(
                NamedLogLevel.ERROR,
                exc_info=exc_info,
                stack_offset=self._level_proxy_method_stack_offset,
            ),
            *args,
            **kwargs,
        )

    #

    @ta.overload
    def critical(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def critical(self, msg: ta.Tuple[ta.Any, ...], **kwargs: ta.Any) -> T:
        ...

    @ta.overload
    def critical(self, msg_fn: LoggingMsgFn, **kwargs: ta.Any) -> T:
        ...

    @ta.final
    def critical(self, *args, **kwargs):
        return self._log(
            CaptureLoggingContextImpl(
                NamedLogLevel.CRITICAL,
                stack_offset=self._level_proxy_method_stack_offset,
            ),
            *args,
            **kwargs,
        )

    #

    @abc.abstractmethod
    def _log(
            self,
            ctx: CaptureLoggingContext,
            msg: ta.Union[str, tuple, LoggingMsgFn],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> T:
        raise NotImplementedError


class Logger(LoggerMetricCollector, AnyLogger[None], Abstract):
    _level_proxy_method_stack_offset: int = 1

    @abc.abstractmethod
    def _log(
            self,
            ctx: CaptureLoggingContext,
            msg: ta.Union[str, tuple, LoggingMsgFn],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> None:
        raise NotImplementedError

    #

    @abc.abstractmethod
    def _metric(self, m: LoggerMetric) -> None:
        raise NotImplementedError


class AsyncLogger(AsyncLoggerMetricCollector, AnyLogger[ta.Awaitable[None]], Abstract):
    _level_proxy_method_stack_offset: int = 0

    @abc.abstractmethod
    def _log(
            self,
            ctx: CaptureLoggingContext,
            msg: ta.Union[str, tuple, LoggingMsgFn],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> ta.Awaitable[None]:
        raise NotImplementedError

    #

    @abc.abstractmethod
    def _metric(self, m: LoggerMetric) -> ta.Awaitable[None]:
        raise NotImplementedError


##


class AnyNopLogger(AnyNopLoggerMetricCollector[T], AnyLogger[T], Abstract):
    @ta.final
    def get_effective_level(self) -> LogLevel:
        return -999


class NopLogger(NopLoggerMetricCollector, AnyNopLogger[None], Logger):
    @ta.final
    def _log(
            self,
            ctx: CaptureLoggingContext,
            msg: ta.Union[str, tuple, LoggingMsgFn],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> None:
        pass


class AsyncNopLogger(AsyncNopLoggerMetricCollector, AnyNopLogger[ta.Awaitable[None]], AsyncLogger):
    @ta.final
    async def _log(
            self,
            ctx: CaptureLoggingContext,
            msg: ta.Union[str, tuple, LoggingMsgFn],
            *args: ta.Any,
            **kwargs: ta.Any,
    ) -> None:
        pass



import typing as ta

from ..json.stream.errors import JsonStreamError
from ..json.stream.utils import DebugJsonStreamValueParser
from ..json.stream.utils import JsonStreamValueParser
from .errors import Json5Error
from .stream import make_machinery


##


def parse(
        buf: str,
        *,
        debug: bool = False,
) -> ta.Any:
    m = make_machinery()

    if debug:
        vc: type[JsonStreamValueParser] = DebugJsonStreamValueParser
    else:
        vc = JsonStreamValueParser

    try:
        return vc.parse_exactly_one_value(m, buf)

    except Json5Error:
        raise

    except JsonStreamError as e:
        raise Json5Error from e


def parse_many(
        buf: str,
        *,
        debug: bool = False,
) -> ta.Iterator[ta.Any]:
    m = make_machinery()

    if debug:
        vc: type[JsonStreamValueParser] = DebugJsonStreamValueParser
    else:
        vc = JsonStreamValueParser

    try:
        yield from vc.parse_values(m, buf)

    except Json5Error:
        raise

    except JsonStreamError as e:
        raise Json5Error from e



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .buffers import (  # noqa
        SyncBufferRelay,
    )

    from .conddeque import (  # noqa
        ConditionDeque,
    )

    from .funcs import (  # noqa
        SyncOnce,
        SyncLazy,
        SyncLazyFn,
    )

    from .latches import (  # noqa
        CountDownLatch,
    )

    from .objectpools import (  # noqa
        ObjectPool,
    )



import typing as ta

from .. import lang
from .keys import Key
from .keys import as_key


if ta.TYPE_CHECKING:
    from .impl import inspect as _inspect
else:
    _inspect = lang.proxy_import('.impl.inspect', __package__)


T = ta.TypeVar('T')


##


class Kwarg(ta.NamedTuple):
    name: str
    key: Key
    has_default: bool

    @classmethod
    def of(
            cls,
            name: str,
            key: Key,
            *,
            has_default: bool = False,
    ) -> 'Kwarg':
        return cls(
            name,
            key,
            has_default,
        )


class KwargsTarget(ta.NamedTuple):
    obj: ta.Any
    kwargs: ta.Sequence[Kwarg]

    @classmethod
    def of(
            cls,
            obj: ta.Any,
            *kws: Kwarg,
            **kwargs: tuple[Key, bool] | Key | ta.Any,
    ) -> 'KwargsTarget':
        kw_kwargs: list[Kwarg] = []
        for n, v in kwargs.items():
            if isinstance(v, tuple):
                kw_k, kw_hd = v
                kw_kwargs.append(Kwarg.of(n, kw_k, has_default=kw_hd))
            else:
                kw_kwargs.append(Kwarg.of(n, as_key(v)))

        return cls(
            obj,
            (*kws, *kw_kwargs),
        )


def tag(obj: T, **kwargs: ta.Any) -> T:
    return _inspect.tag(obj, **kwargs)


def build_kwargs_target(
        obj: ta.Any,
        *,
        skip_args: int = 0,
        skip_kwargs: ta.Iterable[str] | None = None,
        raw_optional: bool = False,
        non_strict: bool = False,
) -> KwargsTarget:
    return _inspect.build_kwargs_target(
        obj,
        skip_args=skip_args,
        skip_kwargs=skip_kwargs,
        raw_optional=raw_optional,
        non_strict=non_strict,
    )


##


def target(**kwargs: ta.Any) -> ta.Callable[[ta.Any], KwargsTarget]:
    def inner(obj: ta.Any) -> KwargsTarget:
        return KwargsTarget.of(obj, **kwargs)
    return inner



import threading
import time
import typing as ta

from .types import ScheduledCallback
from .types import ScheduledCancelledError
from .types import ScheduledTimeoutError
from .types import ScheduleHandle


##


class ScheduleHandleImpl(ScheduleHandle):
    """
    A cancellable handle for a scheduled function.

    Thread-safe. Callbacks run in the runner thread (so keep them short).
    """

    def __init__(self, cv: threading.Condition) -> None:
        super().__init__()

        self._cv = cv
        self._callbacks: list[ScheduledCallback] = []

    _done = False
    _cancelled = False
    _running = False
    _result: ta.Any = None
    _exc: BaseException | None = None

    # status

    def done(self) -> bool:
        with self._cv:
            return self._done

    def cancelled(self) -> bool:
        with self._cv:
            return self._cancelled

    def running(self) -> bool:
        with self._cv:
            return self._running

    # callbacks

    def _invoke_callbacks(self, cbs: ta.Sequence[ScheduledCallback]) -> None:
        for cb in cbs:
            cb(self)

    # completion

    def cancel(self) -> bool:
        with self._cv:
            if self._done or self._running:
                return False
            if self._cancelled:
                return True

            self._cancelled = True
            self._done = True

            cbs = list(self._callbacks)
            self._callbacks.clear()

            self._cv.notify_all()

        # Callbacks run outside lock (in caller thread)
        self._invoke_callbacks(cbs)

        return True

    def result(self, timeout: float | None = None) -> ta.Any:
        with self._cv:
            if not self._wait_done_locked(timeout):
                raise ScheduledTimeoutError('scheduled result timed out')
            if self._cancelled:
                raise ScheduledCancelledError('scheduled task was cancelled')

            if self._exc is not None:
                raise self._exc
            return self._result

    def exception(self, timeout: float | None = None) -> BaseException | None:
        with self._cv:
            if not self._wait_done_locked(timeout):
                raise ScheduledTimeoutError('scheduled result timed out')
            if self._cancelled:
                raise ScheduledCancelledError('scheduled task was cancelled')

            return self._exc

    def add_done_callback(self, fn: ScheduledCallback) -> None:
        call_now = False

        with self._cv:
            if self._done:
                call_now = True
            else:
                self._callbacks.append(fn)

        if call_now:
            fn(self)

    # internal (runner thread)

    def _set_running(self) -> None:
        with self._cv:
            self._running = True

    def _set_result(self, result: ta.Any) -> None:
        with self._cv:
            if self._done:
                return

            self._result = result
            self._exc = None
            self._running = False
            self._done = True

            cbs = list(self._callbacks)
            self._callbacks.clear()

            self._cv.notify_all()

        self._invoke_callbacks(cbs)

    def _set_exception(self, exc: BaseException) -> None:
        with self._cv:
            if self._done:
                return

            self._result = None
            self._exc = exc
            self._running = False
            self._done = True

            cbs = list(self._callbacks)
            self._callbacks.clear()

            self._cv.notify_all()

        self._invoke_callbacks(cbs)

    def _wait_done_locked(self, timeout: float | None) -> bool:
        if self._done:
            return True

        if timeout is None:
            while not self._done:
                self._cv.wait()
            return True  # type: ignore[unreachable]  # ??

        deadline = time.monotonic() + max(0.0, float(timeout))
        while not self._done:
            remaining = deadline - time.monotonic()
            if remaining <= 0:
                return False
            self._cv.wait(timeout=remaining)

        return True  # type: ignore[unreachable]  # ??



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import io
import typing as ta

from ...lite.abstract import Abstract
from ..types import BytesLike
from .direct import DirectByteStreamBufferView
from .errors import NeedMoreDataByteStreamBufferError
from .types import ByteStreamBuffer
from .types import ByteStreamBufferLike
from .types import ByteStreamBufferView
from .types import MutableByteStreamBuffer


BytesOrAwaitableBytes = ta.TypeVar('BytesOrAwaitableBytes', bound=ta.Union[bytes, ta.Awaitable[bytes]])
BytesOrAwaitableBytes_co = ta.TypeVar('BytesOrAwaitableBytes_co', bound=ta.Union[bytes, ta.Awaitable[bytes]], covariant=True)  # noqa

BoolOrAwaitableBool = ta.TypeVar('BoolOrAwaitableBool', bound=ta.Union[bool, ta.Awaitable[bool]])
BoolOrAwaitableBool_co = ta.TypeVar('BoolOrAwaitableBool_co', bound=ta.Union[bool, ta.Awaitable[bool]], covariant=True)


##


BaseByteStreamBufferBytesReaderAdapterPolicy = ta.Literal[  # ta.TypeAlias  # omlish-amalg-typing-no-move
    'raise',
    'return_partial',
    'fill',
]


class BaseByteStreamBufferBytesReaderAdapter(Abstract, ta.Generic[BytesOrAwaitableBytes, BoolOrAwaitableBool]):
    """
    Adapter: ByteStreamBuffer -> file-like buffered reader methods (`read1`, `read`).

    This adapter is policy-driven for how it behaves when insufficient bytes are available. The core buffer is
    intentionally non-blocking; blocking behavior (if desired) must be provided via a `fill` callback that supplies more
    bytes into the buffer.

    `policy`:
      - 'raise': raise `NeedMoreDataByteStreamBufferError` if fewer than `n` bytes are available (for `n >= 0`).
      - 'return_partial': return whatever is available (possibly b'') up to `n`.
      - 'fill': repeatedly call `fill()` until satisfied or until `fill()` signals EOF - effectively blocking mode.

    `fill`:
      - Callable that writes more bytes into the underlying buffer.

    This adapter exists for interop with legacy code that expects file-like objects, but the core design remains
    independent from `io` and blocking semantics.
    """

    DEFAULT_POLICY: ta.Final = 'raise'

    class Filler(ta.Protocol[BoolOrAwaitableBool_co]):
        def __call__(self, n: int, single: bool) -> BoolOrAwaitableBool_co:
            """
            Called when more data needs to be written into the underlying buffer.

            - If `n > 0` then at least `n` bytes are requested - extra bytes will be buffered.
            - If `single` then the user has requested that at most one underlying read operation should be performed.
            - If `n < 1` and `single` then any amount of data is requested.
            - If `n < 1` and not `single` then all remaining data is requested.

            Returns `True` if it made progress or more data may be available, or `False` if no more data will arrive.

            If not `single` returning `True` then at least 1 new byte mut have been written to the buffer (otherwise
            it would infinite loop).

            This function may or may not sleep / loop / block / etc. as it sees fit.
            """

    def __init__(
            self,
            buf: ByteStreamBuffer,
            *,
            policy: ta.Optional[BaseByteStreamBufferBytesReaderAdapterPolicy] = None,
            fill: ta.Optional[Filler[BoolOrAwaitableBool]] = None,
    ) -> None:
        super().__init__()

        self._buf = buf
        if fill is not None:
            if policy is None:
                policy = 'fill'
            elif policy != 'fill':
                raise ValueError('fill callback only valid with policy=fill')
        elif policy is None:
            policy = self.DEFAULT_POLICY
        self._policy = policy
        self._fill = fill

    #

    class RawBytesReader(ta.Protocol[BytesOrAwaitableBytes_co]):
        def read1(self, n: ta.Any) -> BytesOrAwaitableBytes_co: ...

    class BytesReader(RawBytesReader[BytesOrAwaitableBytes_co], ta.Protocol[BytesOrAwaitableBytes_co]):
        def read(self, n: ta.Any) -> BytesOrAwaitableBytes_co: ...

    @classmethod
    @abc.abstractmethod
    def wrap(
            cls,
            obj: RawBytesReader[BytesOrAwaitableBytes],
            buf: MutableByteStreamBuffer,
            *,
            policy: ta.Optional[BaseByteStreamBufferBytesReaderAdapterPolicy] = None,
    ) -> 'BaseByteStreamBufferBytesReaderAdapter[BytesOrAwaitableBytes, BoolOrAwaitableBool]':
        raise NotImplementedError

    #

    @abc.abstractmethod
    def read1(self, n: int = -1, /) -> BytesOrAwaitableBytes:
        raise NotImplementedError

    def _inner_read(self, n: int) -> ta.Union[bytes, ta.Literal['fill']]:
        buf = self._buf
        if (ln := len(buf)) >= n:
            return buf.split_to(n).tobytes()

        pol = self._policy
        if pol == 'return_partial':
            if not ln:
                return b''
            return buf.split_to(ln).tobytes()

        elif pol == 'raise':
            raise NeedMoreDataByteStreamBufferError

        elif pol == 'fill':
            return 'fill'

        else:
            raise RuntimeError(f'invalid policy: {pol!r}')

    @abc.abstractmethod
    def read(self, n: int = -1, /) -> BytesOrAwaitableBytes:
        raise NotImplementedError


#


class ByteStreamBufferBytesReaderAdapter(BaseByteStreamBufferBytesReaderAdapter[bytes, bool]):
    @classmethod
    def wrap(
            cls,
            obj: BaseByteStreamBufferBytesReaderAdapter.RawBytesReader[bytes],
            buf: MutableByteStreamBuffer,
            *,
            policy: ta.Optional[BaseByteStreamBufferBytesReaderAdapterPolicy] = None,
    ) -> 'BaseByteStreamBufferBytesReaderAdapter[bytes, bool]':
        if not hasattr(obj, 'read1'):
            raise TypeError(obj)

        elif hasattr(obj, 'read'):
            def br_fill(n: int, single: bool) -> bool:
                if single:
                    b = obj.read1(n)
                else:
                    b = obj.read(n)
                if not b:
                    return False
                buf.write(b)
                return True

            fill = br_fill

        else:
            def rbr_fill(n: int, single: bool) -> bool:
                b = obj.read1(n)
                if not b:
                    return False
                buf.write(b)
                return True

            fill = rbr_fill

        return cls(
            buf,
            policy=policy,
            fill=fill,
        )

    def read1(self, n: int = -1, /) -> bytes:
        if not n:
            return b''

        buf = self._buf
        if not (ln := len(buf)):
            if (fill := self._fill) is None:
                return b''

            fill(n, True)
            ln = len(buf)

        if not ln:
            return b''

        return buf.split_to(min(n, ln) if n > 0 else ln).tobytes()

    def read(self, n: int = -1, /) -> bytes:
        if n < 0:
            if (fill := self._fill) is not None:
                while fill(-1, False):
                    pass

            buf = self._buf
            return buf.split_to(len(buf)).tobytes()

        if not n:
            return b''

        buf = self._buf
        while True:
            if (out := self._inner_read(n)) != 'fill':
                return out

            ln = len(buf)
            if not self._fill(n - ln, False):  # type: ignore[misc]
                # EOF
                if not (ln := len(buf)):
                    return b''
                return buf.split_to(ln).tobytes()

            if len(buf) == ln:
                raise RuntimeError('fill did not produce data')


#


class ByteStreamBufferAsyncBytesReaderAdapter(BaseByteStreamBufferBytesReaderAdapter[ta.Awaitable[bytes], ta.Awaitable[bool]]):  # noqa
    @classmethod
    def wrap(
            cls,
            obj: BaseByteStreamBufferBytesReaderAdapter.RawBytesReader[ta.Awaitable[bytes]],
            buf: MutableByteStreamBuffer,
            *,
            policy: ta.Optional[BaseByteStreamBufferBytesReaderAdapterPolicy] = None,
    ) -> 'BaseByteStreamBufferBytesReaderAdapter[ta.Awaitable[bytes], ta.Awaitable[bool]]':
        if not hasattr(obj, 'read1'):
            raise TypeError(obj)

        elif hasattr(obj, 'read'):
            async def br_fill(n: int, single: bool) -> bool:
                if single:
                    b = await obj.read1(n)
                else:
                    b = await obj.read(n)
                if not b:
                    return False
                buf.write(b)
                return True

            fill = br_fill

        else:
            async def rbr_fill(n: int, single: bool) -> bool:
                b = await obj.read1(n)
                if not b:
                    return False
                buf.write(b)
                return True

            fill = rbr_fill

        return cls(
            buf,
            policy=policy,
            fill=fill,
        )

    async def read1(self, n: int = -1, /) -> bytes:
        if not n:
            return b''

        buf = self._buf
        if not (ln := len(buf)):
            if (fill := self._fill) is None:
                return b''

            await fill(n, True)
            ln = len(buf)

        if not ln:
            return b''

        return buf.split_to(min(n, ln) if n > 0 else ln).tobytes()

    async def read(self, n: int = -1, /) -> bytes:
        if n < 0:
            if (fill := self._fill) is not None:
                while await fill(-1, False):
                    pass

            buf = self._buf
            return buf.split_to(len(buf)).tobytes()

        if not n:
            return b''

        buf = self._buf
        while True:
            if (out := self._inner_read(n)) != 'fill':
                return out

            ln = len(buf)
            if not await self._fill(n - ln, False):  # type: ignore[misc]
                # EOF
                if not (ln := len(buf)):
                    return b''
                return buf.split_to(ln).tobytes()

            if len(buf) == ln:
                raise RuntimeError('fill did not produce data')


##


class ByteStreamBufferWriterAdapter:
    """
    Adapter: file-like writer sink <- ByteStreamBuffer / bytes-like.

    This is intentionally small and dumb: it exists to bridge into code expecting a `.write(...)` method on an object.

    If given a ByteStreamBufferView-like object, it writes segment-by-segment to avoid materializing copies when the
    sink can accept multiple writes efficiently.
    """

    def __init__(self, f: ta.Any) -> None:
        super().__init__()

        self._f = f

    def write(self, data: ta.Any) -> int:
        f = self._f

        if isinstance(data, (bytes, bytearray, memoryview)):
            b = data.tobytes() if isinstance(data, memoryview) else bytes(data)
            return ta.cast(int, f.write(b))

        if isinstance(data, ByteStreamBufferLike):
            total = 0
            for mv in data.segments():
                total += ta.cast(int, f.write(bytes(mv)))
            return total

        if isinstance(data, ByteStreamBufferView):
            b = data.tobytes()
            return ta.cast(int, f.write(b))

        raise TypeError(data)


##


class BytesIoByteStreamBuffer(MutableByteStreamBuffer):
    """
    ByteStreamBuffer/MutableByteStreamBuffer implementation backed by `io.BytesIO`, using `getbuffer()`.

    This exists primarily for interoperability with code that already produces/consumes `BytesIO`, and to demonstrate
    how `getbuffer()` can expose a non-copying `memoryview` of internal storage.

    Caveat (important):
      - Any exported `memoryview` from `BytesIO.getbuffer()` can pin the BytesIO against resizing.
      - If a caller holds onto a view and we attempt to grow/resize internally, `BytesIO` may raise `BufferError`.

    This backing is therefore best suited for controlled scenarios; it is *not* the recommended default buffer backend
    for applications (segmented/bytearray backends are more predictable).
    """

    def __init__(
            self,
            *,
            compaction_threshold: int = 64 * 1024,
    ) -> None:
        super().__init__()

        self._compaction_threshold = compaction_threshold

        self._bio = io.BytesIO()
        self._rpos = 0

        # reserve/commit state
        self._resv: ta.Optional[bytearray] = None
        self._resv_len = 0

    @property
    def max_size(self) -> ta.Optional[int]:
        return None

    def __len__(self) -> int:
        return self._bio.getbuffer().nbytes - self._rpos

    def peek(self) -> memoryview:
        mv = self._bio.getbuffer()
        return mv[self._rpos:]

    def segments(self) -> ta.Sequence[memoryview]:
        mv = self._bio.getbuffer()
        seg = mv[self._rpos:]
        return (seg,) if len(seg) else ()

    def advance(self, n: int, /) -> None:
        if n < 0 or n > len(self):
            raise ValueError(n)
        self._rpos += n
        # Optional compaction heuristic: if we've consumed a lot, rebuild a smaller BytesIO.
        # This may fail if someone holds a getbuffer() view (BufferError).
        if (
                self._rpos and
                self._rpos >= self._compaction_threshold and
                self._rpos >= (self._bio.getbuffer().nbytes // 2)
        ):
            try:
                remaining = self._bio.getbuffer()[self._rpos:].tobytes()
                self._bio = io.BytesIO(remaining)
                self._rpos = 0
            except BufferError as e:  # noqa
                raise

    def split_to(self, n: int, /) -> ByteStreamBufferView:
        if n < 0 or n > len(self):
            raise ValueError(n)
        mv = self._bio.getbuffer()
        out = mv[self._rpos:self._rpos + n]
        self._rpos += n
        return DirectByteStreamBufferView(out)

    def write(self, data: BytesLike, /) -> None:
        if not data:
            return
        if isinstance(data, memoryview):
            data = data.tobytes()
        try:
            self._bio.seek(0, io.SEEK_END)
            self._bio.write(data)
        except BufferError as e:  # noqa
            raise

    def reserve(self, n: int, /) -> memoryview:
        if n < 0:
            raise ValueError(n)
        if self._resv is not None:
            raise RuntimeError('outstanding reserve')
        b = bytearray(n)
        self._resv = b
        self._resv_len = n
        return memoryview(b)

    def commit(self, n: int, /) -> None:
        if self._resv is None:
            raise RuntimeError('no outstanding reserve')
        if n < 0 or n > self._resv_len:
            raise ValueError(n)
        b = self._resv
        self._resv = None
        self._resv_len = 0
        if n:
            self.write(memoryview(b)[:n])

    def coalesce(self, n: int, /) -> memoryview:
        if n < 0 or n > len(self):
            raise ValueError(n)
        if not n:
            return memoryview(b'')
        # BytesIO is always contiguous, just return the requested slice
        mv = self._bio.getbuffer()
        return mv[self._rpos:self._rpos + n]

    def find(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        # Normalize start/end to valid range
        s, e, _ = slice(start, end, 1).indices(len(self))
        if e < s:
            e = s

        if not sub:
            return s

        # Convert to bytes for searching (memoryview doesn't have find method)
        b = bytes(self._bio.getbuffer())
        idx = b.find(sub, self._rpos + s, self._rpos + e)
        return (idx - self._rpos) if idx >= 0 else -1

    def rfind(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        # Normalize start/end to valid range
        s, e, _ = slice(start, end, 1).indices(len(self))
        if e < s:
            e = s

        if not sub:
            return e

        # Convert to bytes for searching (memoryview doesn't have rfind method)
        b = bytes(self._bio.getbuffer())
        idx = b.rfind(sub, self._rpos + s, self._rpos + e)
        return (idx - self._rpos) if idx >= 0 else -1



import typing as ta

from ... import check
from ... import reflect as rfl
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory


##


class NewtypeMarshalerFactory(MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not isinstance(rty, rfl.NewType):
            return None
        return lambda: ctx.make_marshaler(check.isinstance(rty, rfl.NewType).ty)


class NewtypeUnmarshalerFactory(UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not isinstance(rty, rfl.NewType):
            return None
        return lambda: ctx.make_unmarshaler(check.isinstance(rty, rfl.NewType).ty)



import typing as ta

from ... import reflect as rfl
from ..api.contexts import MarshalContext
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory
from ..api.values import Value


FactoryT = ta.TypeVar('FactoryT', bound=MarshalerFactory | UnmarshalerFactory)
T = ta.TypeVar('T')


##


class _RecursiveFactory(ta.Generic[FactoryT]):
    def __init__(
            self,
            fac: FactoryT,
            prx: ta.Callable[[], tuple[ta.Any, ta.Callable[[ta.Any], None]]],
    ) -> None:
        super().__init__()

        self._fac = fac
        self._prx = prx

        self._dct: dict[rfl.Type, ta.Any] = {}

    def _wrap(self, m, rty):
        def inner():
            try:
                return self._dct[rty]
            except KeyError:
                pass

            p, sp = self._prx()
            self._dct[rty] = p
            try:
                r = m()
                sp(r)
                return r
            finally:
                del self._dct[rty]

        return inner


class _Proxy(ta.Generic[T]):
    __obj: T | None = None

    @property
    def _obj(self) -> T:
        if self.__obj is None:
            raise TypeError('recursive proxy not set')
        return self.__obj

    def _set_obj(self, obj: T) -> None:
        if self.__obj is not None:
            raise TypeError('recursive proxy already set')
        self.__obj = obj

    @classmethod
    def _new(cls) -> tuple[ta.Any, ta.Callable[[ta.Any], None]]:
        return (p := cls()), p._set_obj  # noqa


#


class _ProxyMarshaler(_Proxy[Marshaler], Marshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return self._obj.marshal(ctx, o)


class RecursiveMarshalerFactory(_RecursiveFactory[MarshalerFactory], MarshalerFactory):
    def __init__(self, fac: MarshalerFactory) -> None:
        super().__init__(fac, _ProxyMarshaler._new)  # noqa

    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if (m := self._fac.make_marshaler(ctx, rty)) is None:
            return None
        return self._wrap(m, rty)


#


class _ProxyUnmarshaler(_Proxy[Unmarshaler], Unmarshaler):
    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self._obj.unmarshal(ctx, v)


class RecursiveUnmarshalerFactory(_RecursiveFactory[UnmarshalerFactory], UnmarshalerFactory):
    def __init__(self, fac: UnmarshalerFactory) -> None:
        super().__init__(fac, _ProxyUnmarshaler._new)  # noqa

    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if (m := self._fac.make_unmarshaler(ctx, rty)) is None:
            return None
        return self._wrap(m, rty)



import abc
import typing as ta

from ... import check
from ... import lang
from .base import Op
from .grammars import Grammar
from .grammars import Rule
from .internal import Regex
from .matches import Match
from .matches import furthest_match
from .matches import longest_match
from .ops import CaseInsensitiveStringLiteral
from .ops import Concat
from .ops import Either
from .ops import RangeLiteral
from .ops import Repeat
from .ops import RuleRef
from .ops import StringLiteral


##

class _Parser(lang.Abstract):
    class MaxStepsExceededError(Exception):
        pass

    @abc.abstractmethod
    def iter_parse(self, op: Op, start: int) -> ta.Iterator[Match]:
        raise NotImplementedError


##


_ParserContext: ta.TypeAlias = tuple[Op, int, ta.Optional['_ParserContext']]


class _ParserImpl(_Parser):
    """
    Packrat-ish-but-non-deterministic? 'Chart' parser?:
    - Top-down
    - Recursive descent
    - Memoized
    - Non-deterministic (yields all possible matches by default unless `Either.first_match`)
    """

    def __init__(
            self,
            grammar: Grammar,
            source: str,
            *,
            max_steps: int | None = None,
    ) -> None:
        super().__init__()

        self._grammar = grammar
        self._source = source
        self._max_steps = max_steps

        self._rules = self._grammar._rules  # Noqa

        self._dispatch: dict[type[Op], ta.Callable[[_ParserContext], ta.Iterator[Match]]] = ta.cast(ta.Any, {  # noqa
            StringLiteral: self._iter_parse_string_literal,
            CaseInsensitiveStringLiteral: self._iter_parse_case_insensitive_string_literal,
            RangeLiteral: self._iter_parse_range_literal,
            Concat: self._iter_parse_concat,
            Repeat: self._iter_parse_repeat,
            Either: self._iter_parse_either,
            RuleRef: self._iter_parse_rule_ref,
            Regex: self._iter_parse_regex,
        })

        self._memo: dict[tuple[Op, int], tuple[Match, ...]] = {}
        self._furthest: _ParserImpl._Furthest | None = None

        self._cur_step = 0

    class _Furthest(ta.NamedTuple):
        op: Op
        match: Match
        ctx: _ParserContext

    def _iter_parse_string_literal(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op: StringLiteral
        op, start, _ = ctx  # type: ignore[assignment]

        if start < len(self._source):  # noqa
            source = self._source[start : start + len(op._value)]  # noqa
            if source == op._value:  # noqa
                yield Match(op, start, start + len(source), ())

    def _iter_parse_case_insensitive_string_literal(self, ctx: _ParserContext) -> ta.Iterator[Match]:  # noqa
        op: CaseInsensitiveStringLiteral
        op, start, _ = ctx  # type: ignore[assignment]

        if start < len(self._source):  # noqa
            source = self._source[start : start + len(op._value)].casefold()  # noqa
            if source == op._value:  # noqa
                yield Match(op, start, start + len(source), ())

    def _iter_parse_range_literal(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op: RangeLiteral
        op, start, _ = ctx  # type: ignore[assignment]

        try:
            source = self._source[start]  # noqa
        except IndexError:
            return

        # ranges are always case-sensitive
        if (value := op._value).lo <= source <= value.hi:  # noqa
            yield Match(op, start, start + 1, ())

    def _iter_parse_concat(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op: Concat
        op, start, _ = ctx  # type: ignore[assignment]

        match_tups: list[tuple[Match, ...]] = [()]

        i = 0
        for cp in op._children:  # noqa
            next_match_tups: list[tuple[Match, ...]] = []

            for mt in match_tups:
                for cm in self._iter_parse((cp, mt[-1].end if mt else start, ctx)):
                    next_match_tups.append((*mt, cm))
                    i += 1

            if not next_match_tups:
                return

            match_tups = next_match_tups

        if not i:
            return

        for mt in sorted(match_tups, key=len, reverse=True):
            yield Match(op, start, mt[-1].end if mt else start, mt)

    def _iter_parse_repeat(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op: Repeat
        op, start, _ = ctx  # type: ignore[assignment]

        # Map from (repetition_count, end_position) to longest match tuple
        matches_by_count_pos: dict[tuple[int, int], tuple[Match, ...]] = {(0, start): ()}
        max_end_by_count: dict[int, int] = {0: start}

        i = 0
        while True:
            if op._times.max is not None and i == op._times.max:  # noqa
                break

            if self._max_steps is not None and self._cur_step > self._max_steps:
                raise _Parser.MaxStepsExceededError(self._cur_step)
            self._cur_step += 1

            next_matches: dict[tuple[int, int], tuple[Match, ...]] = {}
            next_max_end = max_end_by_count.get(i, -1)

            for (count, end_pos), mt in matches_by_count_pos.items():
                if count != i:
                    continue

                for cm in self._iter_parse((op._child, end_pos, ctx)):  # noqa
                    next_mt = (*mt, cm)
                    next_key = (i + 1, cm.end)

                    # Keep only the longest match tuple for each (count, position)
                    if next_key not in next_matches or len(next_mt) > len(next_matches[next_key]):
                        next_matches[next_key] = next_mt
                        if cm.end > next_max_end:
                            next_max_end = cm.end

            if not next_matches:
                break

            # Check if we made progress (reached new positions)
            if next_max_end <= max_end_by_count.get(i, -1):
                break

            i += 1
            matches_by_count_pos.update(next_matches)
            max_end_by_count[i] = next_max_end

        if i < op._times.min:  # noqa
            return

        # Collect valid matches and sort by (end_position, repetition_count) descending
        valid_matches: list[tuple[int, int, tuple[Match, ...]]] = []
        for (count, end_pos), mt in matches_by_count_pos.items():
            if op._times.min <= count <= (op._times.max if op._times.max is not None else i):  # noqa
                valid_matches.append((end_pos, count, mt))

        for end_pos, _, mt in sorted(valid_matches, key=lambda x: (x[0], x[1]), reverse=True):
            yield Match(op, start, end_pos, mt)

    def _iter_parse_either(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op: Either
        op, start, _ = ctx  # type: ignore[assignment]

        for cp in op._children:  # noqa
            found = False

            for cm in self._iter_parse((cp, start, ctx)):
                found = True
                yield Match(op, start, cm.end, (cm,))

            if found and op._first_match:  # noqa
                return

    def _iter_parse_rule_ref(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op: RuleRef
        op, start, _ = ctx  # type: ignore[assignment]

        cp = self._rules._rules_by_name_f[op._name_f].op  # noqa
        for cm in self._iter_parse((cp, start, ctx)):
            yield Match(op, cm.start, cm.end, (cm,))

    def _iter_parse_regex(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op: Regex
        op, start, _ = ctx  # type: ignore[assignment]

        if (m := op._pat.match(self._source, start)) is not None:  # noqa
            yield Match(op, start, m.end(), ())

    def _iter_parse(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op, start, _ = ctx

        if (key := (op, start)) in self._memo:
            yield from self._memo[key]
            return

        if self._max_steps is not None and self._cur_step >= self._max_steps:
            raise _Parser.MaxStepsExceededError(self._cur_step)
        self._cur_step += 1

        matches = tuple(self._dispatch[op.__class__](ctx))
        self._memo[key] = matches

        if matches and (fm := furthest_match(matches)) is not None:
            if self._furthest is None or fm.end > self._furthest.match.end:
                self._furthest = _ParserImpl._Furthest(op, fm, ctx)

        yield from matches

    def iter_parse(self, op: Op, start: int) -> ta.Iterator[Match]:
        return self._iter_parse((op, start, None))

##


class _DebugParserImpl(_ParserImpl):
    def __init__(
            self,
            grammar: Grammar,
            source: str,
            level: int = 1,
            *,
            write: ta.Callable[[str], None] | None = None,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__(grammar, source, **kwargs)

        self._level = level
        if write is None:
            write = print
        self._write = write

        self._op_strs: dict[Op, str] = {}

    def _op_str(self, op: Op) -> str:
        try:
            return self._op_strs[op]
        except KeyError:
            pass
        ps = self._op_strs[op] = str(op)
        return ps

    _depth: int = 0

    def _iter_parse(self, ctx: _ParserContext) -> ta.Iterator[Match]:
        op, start, _ = ctx

        if self._level < 2 and not isinstance(op, RuleRef):
            yield from super()._iter_parse(ctx)
            return

        ws = f'{"  " * self._depth} '

        if self._level < 2:
            ps = check.isinstance(op, RuleRef).name
        else:
            ps = self._op_str(op)
        body = f'{start}:{self._source[start] if start < len(self._source) else ""!r} {ps}'

        if self._level > 2:
            self._write(f'{ws}+ {body}')
        else:
            self._write(f'{ws}{body}')

        try:
            self._depth += 1

            for m in super()._iter_parse(ctx):  # noqa
                if self._level > 3:
                    self._write(f'{ws}! {m.start}-{m.end}:{self._source[m.start:m.end]!r}')

                yield m

        finally:
            self._depth -= 1

            if self._level > 3:
                self._write(f'{ws}- {body}')


#


def _make_parser(
        grammar: Grammar,
        source: str,
        *,
        debug: int = 0,
        max_steps: int | None = None,
) -> _Parser:
    if debug:
        return _DebugParserImpl(
            grammar,
            source,
            max_steps=max_steps,
            level=debug,
        )
    else:
        return _ParserImpl(
            grammar,
            source,
            max_steps=max_steps,
        )


##


def iter_parse(
        obj: Grammar | Rule | Op,
        src: str,
        *,
        root: str | None = None,
        start: int = 0,
        debug: int = 0,
        max_steps: int | None = None,
) -> ta.Iterator[Match]:
    if isinstance(obj, Grammar):
        gram = obj
    elif isinstance(obj, Rule):
        check.none(root)
        gram = Grammar(obj, root=obj)
    elif isinstance(obj, Op):
        check.none(root)
        gram = Grammar(Rule('root', obj), root='root')
    else:
        raise TypeError(obj)

    return gram.iter_parse(
        src,
        root,
        start=start,
        debug=debug,
        max_steps=max_steps,
    )


def parse(
        obj: Grammar | Rule | Op,
        src: str,
        *,
        root: str | None = None,
        start: int = 0,
        debug: int = 0,
        max_steps: int | None = None,
) -> Match | None:
    return longest_match(iter_parse(
        obj,
        src,
        root=root,
        start=start,
        debug=debug,
        max_steps=max_steps,
    ))



from .. import dataclasses as _dc


_dc.init_package(
    globals(),
    codegen=True,
)


##


from .base import (  # noqa
    Lifecycle,
    AsyncLifecycle,
    AnyLifecycle,
    ANY_LIFECYCLE_TYPES,

    sync_to_async_lifecycle,
    as_async_lifecycle,
    async_to_sync_lifecycle,
    as_sync_lifecycle,

    CallbackLifecycle,
    CallbackAsyncLifecycle,
)

from .contextmanagers import (  # noqa
    ContextManagerLifecycle,
    AsyncContextManagerLifecycle,

    LifecycleContextManager,
    AsyncLifecycleContextManager,

    lifecycle_context_manage,
    async_lifecycle_context_manage,
)

from .controller import (  # noqa
    LifecycleController,
    AsyncLifecycleController,
    AnyLifecycleController,
    ANY_LIFECYCLE_CONTROLLER_TYPES,
)

from .listeners import (  # noqa
    LifecycleListener,
    AsyncLifecycleListener,
    AnyLifecycleListener,
    ANY_LIFECYCLE_LISTENER_TYPES,
)

from .managed import (  # noqa
    LifecycleManaged,
    AsyncLifecycleManaged,
    AnyLifecycleManaged,
    ANY_LIFECYCLE_MANAGED_TYPES,
)

from .manager import (  # noqa
    LifecycleManagerEntry,
    LifecycleManager,
    AsyncLifecycleManager,
)

from .states import (  # noqa
    LifecycleStateError,
    LifecycleState,
    LifecycleStates,
)

from .transitions import (  # noqa
    LifecycleTransition,
    LifecycleTransitions,
)

from .unwrap import (  # noqa
    unwrap_lifecycle,
    unwrap_async_lifecycle,
    unwrap_any_lifecycle,
)


##


from .. import lang as _lang  # noqa


with _lang.auto_proxy_init(globals()):
    ##

    from .injection import (  # noqa
        bind_lifecycle_registrar,
        bind_async_lifecycle_registrar,

        bind_managed_lifecycle_manager,
        bind_async_managed_lifecycle_manager,
    )


##


NEW = LifecycleStates.NEW

CONSTRUCTING = LifecycleStates.CONSTRUCTING
FAILED_CONSTRUCTING = LifecycleStates.FAILED_CONSTRUCTING
CONSTRUCTED = LifecycleStates.CONSTRUCTED

STARTING = LifecycleStates.STARTING
FAILED_STARTING = LifecycleStates.FAILED_STARTING
STARTED = LifecycleStates.STARTED

STOPPING = LifecycleStates.STOPPING
FAILED_STOPPING = LifecycleStates.FAILED_STOPPING
STOPPED = LifecycleStates.STOPPED

DESTROYING = LifecycleStates.DESTROYING
FAILED_DESTROYING = LifecycleStates.FAILED_DESTROYING
DESTROYED = LifecycleStates.DESTROYED



# ruff: noqa: UP043 UP045
# @omlish-lite
import http.client
import typing as ta
import urllib.error
import urllib.request

from ...io.streams.adapters import ByteStreamBufferBytesReaderAdapter
from ...io.streams.segmented import SegmentedByteStreamBuffer
from ...lite.check import check
from ..headers import HttpHeaders
from .base import DEFAULT_HTTP_CLIENT_ENCODING
from .base import HttpClientContext
from .base import HttpClientError
from .base import HttpClientRequest
from .sync import HttpClient
from .sync import StreamHttpClientResponse


##


class UrllibHttpClient(HttpClient):
    def _build_request(self, req: HttpClientRequest) -> urllib.request.Request:
        d: ta.Any
        if (d := req.data) is not None:
            if isinstance(d, str):
                d = d.encode(DEFAULT_HTTP_CLIENT_ENCODING)

        # urllib headers are dumb dicts [1], and keys *must* be strings or it will automatically add problematic default
        # headers because it doesn't see string keys in its header dict [2]. frustratingly it has no problem accepting
        # bytes values though [3].
        # [1]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/urllib/request.py#L319-L325  # noqa
        # [2]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/urllib/request.py#L1276-L1279  # noqa
        # [3]: https://github.com/python/cpython/blob/232b303e4ca47892f544294bf42e31dc34f0ec72/Lib/http/client.py#L1300-L1301  # noqa
        h: dict[str, str] = {}
        if hs := req.headers_:
            for k, vs in hs.items():
                h[k] = check.single(vs)

        return urllib.request.Request(  # noqa
            req.url,
            method=req.method_or_default,
            headers=h,
            data=d,
        )

    def _stream_request(self, ctx: HttpClientContext, req: HttpClientRequest) -> StreamHttpClientResponse:
        try:
            resp = urllib.request.urlopen(  # noqa
                self._build_request(req),
                timeout=req.timeout_s,
            )

        except urllib.error.HTTPError as e:
            try:
                return StreamHttpClientResponse(
                    status=e.code,
                    headers=HttpHeaders(e.headers.items()),
                    request=req,
                    underlying=e,
                    _closer=e.close,
                )

            except Exception:
                e.close()
                raise

        except (urllib.error.URLError, http.client.HTTPException) as e:
            raise HttpClientError from e

        try:
            return StreamHttpClientResponse(
                status=resp.status,
                headers=HttpHeaders(resp.headers.items()),
                request=req,
                underlying=resp,
                # urllib responses do *not* guarantee returning exactly `n` bytes from `.read(n)`, so we need to buffer.
                _stream=ByteStreamBufferBytesReaderAdapter.wrap(
                    resp,
                    SegmentedByteStreamBuffer(chunk_size=16 * 1024),
                ),
                _closer=resp.close,
            )

        except Exception:  # noqa
            resp.close()
            raise



# ruff: noqa: UP006 UP007 UP043 UP045
# @omlish-lite
"""
TODO:
 - validating mode?
"""
import abc
import dataclasses as dc
import typing as ta

from ...io.pipelines.bytes.buffering import InboundBytesBufferingIoPipelineHandler
from ...io.pipelines.core import IoPipelineHandler
from ...io.pipelines.core import IoPipelineHandlerContext
from ...io.pipelines.core import IoPipelineMessages
from ...io.pipelines.handlers.decoders import MessageToMessageDecoderIoPipelineHandler
from ...io.streams.errors import FrameTooLargeByteStreamBufferError
from ...io.streams.segmented import SegmentedByteStreamBuffer
from ...io.streams.types import MutableByteStreamBuffer
from ...io.streams.utils import ByteStreamBuffers
from ...io.streams.utils import CanByteStreamBuffer
from ...io.types import BytesLike
from ...lite.abstract import Abstract
from .bodymodes import IoPipelineHttpBodyMode
from .bodymodes import IoPipelineHttpBodyModeError
from .objects import IoPipelineHttpMessageHead
from .objects import IoPipelineHttpMessageObjects


##


@dc.dataclass(frozen=True)
class IoPipelineHttpAggregationConfig:
    DEFAULT: ta.ClassVar['IoPipelineHttpAggregationConfig']

    @dc.dataclass(frozen=True)
    class BufferConfig:
        max_size: ta.Optional[int]
        chunk_size: int

    body_buffer: BufferConfig = BufferConfig(max_size=64 * 1024, chunk_size=64 * 1024)


IoPipelineHttpAggregationConfig.DEFAULT = IoPipelineHttpAggregationConfig()


#


class IoIoPipelineHttpObjectAggregator(
    IoPipelineHttpMessageObjects,
    IoPipelineHandler,
    Abstract,
):
    def __init__(
            self,
            *,
            config: IoPipelineHttpAggregationConfig = IoPipelineHttpAggregationConfig.DEFAULT,
            enabled: ta.Union[bool, ta.Literal['unless_chunked']] = True,
    ) -> None:
        super().__init__()

        self._config = config
        self._enabled = enabled

        self._handled_types: ta.Tuple[type, ...] = (
            self._head_type,
            self._chunk_type,
            self._end_chunk_type,
            self._last_chunk_type,
            self._chunked_trailers_type,
            self._body_data_type,
            self._end_type,
            self._aborted_type,
            self._final_type,
        )

        self._state: IoIoPipelineHttpObjectAggregator._State = self._init_state()

    @property
    def enabled(self) -> ta.Union[bool, ta.Literal['unless_chunked']]:
        return self._enabled

    def set_enabled(self, enabled: ta.Union[bool, ta.Literal['unless_chunked']]) -> None:
        self._enabled = enabled

    #

    @property
    @abc.abstractmethod
    def _if_content_length_missing(self) -> ta.Literal['empty', 'eof']:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def _final_type(self) -> type:
        raise NotImplementedError

    #

    def buffered_bytes(self) -> int:
        if (buf := self._state.buf) is None:
            return 0
        return len(buf)

    #

    def _should_handle(self, msg: ta.Any) -> bool:
        return isinstance(msg, self._handled_types)

    def _handle(
            self,
            ctx: IoPipelineHandlerContext,
            msg: ta.Any,
            out: ta.List[ta.Any],
    ) -> None:
        if isinstance(msg, self._aborted_type):
            self._state = self._AbortedState(self)
            out.append(msg)
            return

        while True:
            if (ret := self._state.handle(ctx, msg, out)) is None:
                return

            self._state, next_msg = ret

            if next_msg is None:
                return

            msg = next_msg

    #

    def _init_state(self) -> '_State':
        if self._enabled is True:
            return self._HeadState(self)
        elif self._enabled == 'unless_chunked':
            return self._UnlessChunkedHeadState(self)
        else:
            return self._DisabledHeadState(self)

    #

    class _State(Abstract):
        def __init__(self, a: 'IoIoPipelineHttpObjectAggregator') -> None:
            super().__init__()

            self._a = a

        @property
        def buf(self) -> ta.Optional[MutableByteStreamBuffer]:
            return None

        def _abort(
                self,
                out: ta.List[ta.Any],
                reason: ta.Union[str, BaseException],
                out_msg: ta.Optional[ta.Any] = None,
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            nxt_state = self._a._AbortedState(self._a)  # noqa
            out.append(self._a._make_aborted(reason))  # noqa
            if out_msg is not None:
                out.append(out_msg)
            return (nxt_state, None)

        @abc.abstractmethod
        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            raise NotImplementedError

    #

    class _HeadState(_State):
        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            if isinstance(msg, self._a._head_type):  # noqa
                try:
                    te = IoPipelineHttpBodyMode.select(
                        msg.headers,
                        if_length_missing=self._a._if_content_length_missing,  # noqa
                    )
                except IoPipelineHttpBodyModeError as e:
                    return self._abort(out, f'Invalid Transfer-Encoding: {e.reason}')

                if te.mode == 'empty':
                    return (self._a._EndState(self._a, msg, b''), None)  # noqa

                if (
                        te.length is not None and
                        (max_body := self._a._config.body_buffer.max_size) is not None and  # noqa
                        te.length > max_body
                ):
                    return self._abort(out, FrameTooLargeByteStreamBufferError('aggregation body exceeded max_body'))

                return (self._a._BodyState(self._a, msg), None)  # noqa

            elif isinstance(msg, self._a._final_type):  # noqa
                out.append(msg)
                return None

            else:
                raise TypeError(f'unexpected message type: {type(msg)}')

    #

    class _BodyState(_State):
        def __init__(
                self,
                a: 'IoIoPipelineHttpObjectAggregator',
                head: IoPipelineHttpMessageHead,
        ) -> None:
            super().__init__(a)

            self._head = head

        _buf: ta.Optional[MutableByteStreamBuffer] = None

        @property
        def buf(self) -> ta.Optional[MutableByteStreamBuffer]:
            return self._buf

        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            if isinstance(msg, self._a._body_data_type):  # noqa
                if (buf := self._buf) is None:
                    buf = self._buf = SegmentedByteStreamBuffer(
                        max_size=self._a._config.body_buffer.max_size,  # noqa
                        chunk_size=self._a._config.body_buffer.chunk_size,  # noqa
                    )

                for mv in ByteStreamBuffers.iter_segments(msg.data):
                    buf.write(mv)

                return None

            elif isinstance(msg, (
                    self._a._chunk_type,  # noqa
                    self._a._end_chunk_type,  # noqa
                    self._a._last_chunk_type,  # noqa
                    self._a._chunked_trailers_type,  # noqa
            )):
                return None

            elif isinstance(msg, self._a._end_type):  # noqa
                body: CanByteStreamBuffer
                if (buf := self._buf) is not None:
                    body = buf.coalesce(len(buf))
                else:
                    body = b''

                full = self._a._make_full(self._head, body)  # noqa
                out.append(full)
                return (self._a._init_state(), None)  # noqa

            elif isinstance(msg, self._a._final_type):  # noqa
                return self._abort(out, 'incomplete message body', msg)

            else:
                raise TypeError(f'unexpected message type: {type(msg)}')

    #

    class _EndState(_State):
        def __init__(
                self,
                a: 'IoIoPipelineHttpObjectAggregator',
                head: IoPipelineHttpMessageHead,
                body: BytesLike,
        ) -> None:
            super().__init__(a)

            self._head = head
            self._body = body

        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            if isinstance(msg, self._a._end_type):  # noqa
                full = self._a._make_full(self._head, self._body)  # noqa
                out.append(full)
                return (self._a._init_state(), None)  # noqa

            elif isinstance(msg, self._a._final_type):  # noqa
                return self._abort(out, 'incomplete message sequence', msg)

            else:
                raise TypeError(f'unexpected message type: {type(msg)}')

    #

    class _UnlessChunkedHeadState(_State):
        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            if isinstance(msg, self._a._head_type):  # noqa
                try:
                    te = IoPipelineHttpBodyMode.select(
                        msg.headers,
                        if_length_missing=self._a._if_content_length_missing,  # noqa
                    )
                except IoPipelineHttpBodyModeError as e:
                    return self._abort(out, f'Invalid Transfer-Encoding: {e.reason}')

                if te.mode != 'chunked':
                    return (self._a._HeadState(self._a), msg)  # noqa
                else:
                    out.append(msg)
                    return (self._a._ReInitEndState(self._a), None)  # noqa

            out.append(msg)
            return None

    class _DisabledHeadState(_State):
        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            out.append(msg)

            if isinstance(msg, self._a._head_type):  # noqa
                return (self._a._ReInitEndState(self._a), None)  # noqa

            return None

    class _ReInitEndState(_State):
        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            out.append(msg)

            if isinstance(msg, self._a._end_type):  # noqa
                return (self._a._init_state(), None)  # noqa

            return None

    #

    class _AbortedState(_State):
        def handle(
                self,
                ctx: IoPipelineHandlerContext,
                msg: ta.Any,
                out: ta.List[ta.Any],
        ) -> ta.Optional[ta.Tuple['IoIoPipelineHttpObjectAggregator._State', ta.Optional[ta.Any]]]:
            if isinstance(msg, IoPipelineMessages.MustPropagate):
                out.append(msg)
                return None

            raise NotImplementedError


#


class IoPipelineHttpObjectAggregatorDecoder(
    InboundBytesBufferingIoPipelineHandler,
    MessageToMessageDecoderIoPipelineHandler,
    IoIoPipelineHttpObjectAggregator,
    Abstract,
):
    _final_type: ta.Final[type] = IoPipelineMessages.FinalInput

    #

    def inbound_buffered_bytes(self) -> int:
        return self.buffered_bytes()

    #

    def _should_decode(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> bool:
        return self._should_handle(msg)

    def _decode(
            self,
            ctx: IoPipelineHandlerContext,
            msg: ta.Any,
            out: ta.List[ta.Any],
    ) -> None:
        self._handle(ctx, msg, out)



import threading
import typing as ta

from .. import lang
from .api.configs import Config
from .api.configs import ConfigRegistry
from .api.types import MarshalerFactory
from .api.types import Marshaling
from .api.types import UnmarshalerFactory
from .api.values import Value
from .factories.moduleimport.api import ModuleImport


if ta.TYPE_CHECKING:
    from . import standard
else:
    standard = lang.proxy_import('.standard', __package__)


T = ta.TypeVar('T')


##


_GLOBAL_LOCK = threading.RLock()


@lang.cached_function(lock=_GLOBAL_LOCK)
def global_config_registry() -> ConfigRegistry:
    return ConfigRegistry(lock=_GLOBAL_LOCK)


@lang.cached_function(lock=_GLOBAL_LOCK)
def global_marshaler_factory() -> MarshalerFactory:
    return standard.new_standard_marshaler_factory()


@lang.cached_function(lock=_GLOBAL_LOCK)
def global_unmarshaler_factory() -> UnmarshalerFactory:
    return standard.new_standard_unmarshaler_factory()


class _GlobalMarshaling(Marshaling, lang.Final):
    def config_registry(self) -> ConfigRegistry:
        return global_config_registry()

    def marshaler_factory(self) -> MarshalerFactory:
        return global_marshaler_factory()

    def unmarshaler_factory(self) -> UnmarshalerFactory:
        return global_unmarshaler_factory()


@lang.cached_function(lock=_GLOBAL_LOCK)
def global_marshaling() -> Marshaling:
    return _GlobalMarshaling()


##


def marshal(obj: ta.Any, ty: ta.Any | None = None, **kwargs: ta.Any) -> Value:
    return global_marshaling().marshal(obj, ty, **kwargs)


@ta.overload
def unmarshal(v: Value, ty: type[T], **kwargs: ta.Any) -> T:
    ...


@ta.overload
def unmarshal(v: Value, ty: ta.Any, **kwargs: ta.Any) -> ta.Any:
    ...


def unmarshal(v, ty, **kwargs):
    return global_marshaling().unmarshal(v, ty, **kwargs)


##


def register_global_config(
        key: ta.Any,
        *items: Config,
        identity: bool = False,
) -> None:
    global_config_registry().register(
        key,
        *items,
        identity=identity,
    )


def register_global_module_import(
        name: str,
        package: str | None = None,
) -> None:
    global_config_registry().register(
        None,
        ModuleImport(name, package),
    )



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .accessor import (  # noqa
        TypedValuesAccessor,
    )

    from .collection import (  # noqa
        DuplicateUniqueTypedValueError,

        TypedValues,

        collect,
        as_collection,
    )

    from .consumer import (  # noqa
        UnconsumedTypedValuesError,

        TypedValuesConsumer,

        consume,
    )

    from .generic import (  # noqa
        TypedValueGeneric,
    )

    from .holder import (  # noqa
        TypedValueHolder,
    )

    from .of_ import (  # noqa
        of,
    )

    from .reflect import (  # noqa
        reflect_typed_values_impls,
    )

    from .values import (  # noqa
        TypedValue,

        UniqueTypedValue,

        ScalarTypedValue,

        UniqueScalarTypedValue,
    )


##


from .. import marshal as _msh


_msh.register_global_module_import('.marshal', __package__)



from .api import (  # noqa
    AsyncioAsynclite as All,
)

from .base import (  # noqa
    AsyncioAsyncliteObject as Object,
    AsyncioAsyncliteApi as Api,
)

from .events import (  # noqa
    AsyncioAsyncliteEvent as Event,
    AsyncioAsyncliteEvents as Events,
)

from .identities import (  # noqa
    AsyncioAsyncliteIdentities as Identities,
)

from .locks import (  # noqa
    AsyncioAsyncliteLock as Lock,
    AsyncioAsyncliteLocks as Locks,
)

from .queues import (  # noqa
    AsyncioAsyncliteQueue as Queue,
    AsyncioAsyncliteQueues as Queues,
)

from .semaphores import (  # noqa
    AsyncioAsyncliteSemaphore as Semaphore,
    AsyncioAsyncliteSemaphores as Semaphores,
)

from .sleeps import (  # noqa
    AsyncioAsyncliteSleeps as Sleeps,
)



# ruff: noqa: UP007 UP045
# @omlish-lite
import dataclasses as dc
import http
import typing as ta

from ...io.types import BytesLike
from ...lite.abstract import Abstract
from ...lite.check import check
from ...lite.dataclasses import install_dataclass_kw_only_init
from ..headers import HttpHeaders
from ..parsing import ParsedHttpMessage
from ..versions import HttpVersion
from ..versions import HttpVersions
from .objects import FullIoPipelineHttpMessage
from .objects import IoPipelineHttpMessageAborted
from .objects import IoPipelineHttpMessageBodyData
from .objects import IoPipelineHttpMessageChunk
from .objects import IoPipelineHttpMessageChunkedTrailers
from .objects import IoPipelineHttpMessageEnd
from .objects import IoPipelineHttpMessageEndChunk
from .objects import IoPipelineHttpMessageHead
from .objects import IoPipelineHttpMessageLastChunk
from .objects import IoPipelineHttpMessageObject
from .objects import IoPipelineHttpMessageObjects


##


class IoPipelineHttpResponseObject(IoPipelineHttpMessageObject, Abstract):
    pass


#


@ta.final
@install_dataclass_kw_only_init()
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseHead(IoPipelineHttpMessageHead, IoPipelineHttpResponseObject):
    status: int
    reason: str

    headers: HttpHeaders
    parsed: ta.Optional[ParsedHttpMessage] = None

    version: HttpVersion = HttpVersions.HTTP_1_1

    @staticmethod
    def get_reason_phrase(code: int) -> str:
        try:
            return http.HTTPStatus(code).phrase
        except ValueError:
            return ''


#


@ta.final
@dc.dataclass(frozen=True)
class FullIoPipelineHttpResponse(FullIoPipelineHttpMessage, IoPipelineHttpResponseObject):
    head: IoPipelineHttpResponseHead
    body: BytesLike

    @classmethod
    def simple(
            cls,
            *,
            version: HttpVersion = HttpVersions.HTTP_1_1,
            status: int = 200,
            reason: ta.Optional[str] = None,

            content_type: str = 'text/plain; charset=utf-8',
            body: bytes = b'',
            connection: str = 'close',

            headers: ta.Optional[ta.Mapping[str, str]] = None,
    ):
        return cls(
            head=IoPipelineHttpResponseHead(
                version=version,
                status=status,
                reason=IoPipelineHttpResponseHead.get_reason_phrase(status) if reason is None else reason,
                headers=HttpHeaders([
                    ('Content-Type', content_type),
                    ('Content-Length', str(len(body))),
                    ('Connection', connection),
                    *(headers.items() if headers else []),
                ]),
            ),
            body=body,
        )


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseChunk(IoPipelineHttpMessageChunk, IoPipelineHttpResponseObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseEndChunk(IoPipelineHttpMessageEndChunk, IoPipelineHttpResponseObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseLastChunk(IoPipelineHttpMessageLastChunk, IoPipelineHttpResponseObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseChunkedTrailers(IoPipelineHttpMessageChunkedTrailers, IoPipelineHttpResponseObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseBodyData(IoPipelineHttpMessageBodyData, IoPipelineHttpResponseObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseEnd(IoPipelineHttpMessageEnd, IoPipelineHttpResponseObject):
    pass


#


@ta.final
@dc.dataclass(frozen=True)
class IoPipelineHttpResponseAborted(IoPipelineHttpMessageAborted, IoPipelineHttpResponseObject):
    pass


##


class IoPipelineHttpResponseObjects(IoPipelineHttpMessageObjects):
    _head_type: ta.Final = IoPipelineHttpResponseHead

    def _make_head(self, parsed: ParsedHttpMessage) -> ta.Any:
        status = check.not_none(parsed.status_line)

        return IoPipelineHttpResponseHead(
            version=status.http_version,
            status=status.status_code,
            reason=status.reason_phrase,
            headers=HttpHeaders(parsed.headers.entries),
            parsed=parsed,
        )

    #

    _full_type: ta.Final = FullIoPipelineHttpResponse

    def _make_full(self, head: IoPipelineHttpMessageHead, body: BytesLike) -> FullIoPipelineHttpResponse:
        return FullIoPipelineHttpResponse(check.isinstance(head, IoPipelineHttpResponseHead), body)

    #

    _chunk_type: ta.Final = IoPipelineHttpResponseChunk

    def _make_chunk(self, size: int) -> IoPipelineHttpResponseChunk:
        return IoPipelineHttpResponseChunk(size)

    #

    _end_chunk_type: ta.Final = IoPipelineHttpResponseEndChunk

    def _make_end_chunk(self) -> IoPipelineHttpResponseEndChunk:
        return IoPipelineHttpResponseEndChunk()

    #

    _last_chunk_type: ta.Final = IoPipelineHttpResponseLastChunk

    def _make_last_chunk(self) -> IoPipelineHttpResponseLastChunk:
        return IoPipelineHttpResponseLastChunk()

    #

    _chunked_trailers_type: ta.Final = IoPipelineHttpResponseChunkedTrailers

    def _make_chunked_trailers(self) -> IoPipelineHttpResponseChunkedTrailers:
        return IoPipelineHttpResponseChunkedTrailers()

    #

    _body_data_type: ta.Final = IoPipelineHttpResponseBodyData

    def _make_body_data(self, data: BytesLike) -> IoPipelineHttpResponseBodyData:
        return IoPipelineHttpResponseBodyData(data)

    #

    _end_type: ta.Final = IoPipelineHttpResponseEnd

    def _make_end(self) -> IoPipelineHttpResponseEnd:
        return IoPipelineHttpResponseEnd()

    #

    _aborted_type: ta.Final = IoPipelineHttpResponseAborted

    def _make_aborted(self, reason: ta.Union[str, BaseException]) -> IoPipelineHttpResponseAborted:
        return IoPipelineHttpResponseAborted(reason)



import typing as ta

from ... import check
from ... import lang
from .asquery import as_query
from .core import AsyncRows
from .core import Rows
from .queriers import AsyncQuerier
from .queriers import Querier
from .queries import QueryMode
from .rows import Row


##


def _sync_async_query_func(sync_fn, async_fn):
    def outer(fn):  # noqa
        def inner(querier, *args, **kwargs):
            if isinstance(querier, AsyncQuerier):
                return async_fn(querier, *args, **kwargs)

            elif isinstance(querier, Querier):
                return sync_fn(querier, *args, **kwargs)

            else:
                raise TypeError(querier)

        return inner

    return outer


##


def sync_exec(  # noqa
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> None:
    q = as_query(
        obj,
        *args,
        mode=QueryMode.EXEC,
        adapter=querier.adapter,
    )

    with querier.query(q):
        pass


async def async_exec(  # noqa
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> None:
    q = as_query(
        obj,
        *args,
        mode=QueryMode.EXEC,
        adapter=querier.adapter,
    )

    async with querier.query(q):
        pass


@ta.overload
def exec(  # noqa
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> None:
    ...


@ta.overload
def exec(  # noqa
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[None]:
    ...


@_sync_async_query_func(sync_exec, async_exec)
def exec(*args, **kwargs):  # noqa
    raise RuntimeError


##


def sync_query(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.ContextManager['Rows']:
    q = as_query(
        obj,
        *args,
        mode=QueryMode.QUERY,
        adapter=querier.adapter,
    )

    return querier.query(q)


def async_query(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.AsyncContextManager['AsyncRows']:
    q = as_query(
        obj,
        *args,
        mode=QueryMode.QUERY,
        adapter=querier.adapter,
    )

    return querier.query(q)


@ta.overload
def query(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.ContextManager['Rows']:
    ...


@ta.overload
def query(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.AsyncContextManager['Rows']:
    ...


@_sync_async_query_func(sync_query, async_query)
def query(*args, **kwargs):
    raise RuntimeError


##


def sync_query_all(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> list[Row]:
    with sync_query(querier, obj, *args) as rows:
        return list(rows)


async def async_query_all(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> list[Row]:
    async with async_query(querier, obj, *args) as rows:
        return await lang.async_list(rows)


@ta.overload
def query_all(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> list[Row]:
    ...


@ta.overload
def query_all(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[list[Row]]:
    ...


@_sync_async_query_func(sync_query_all, async_query_all)
def query_all(*args, **kwargs):
    raise RuntimeError


#


def sync_query_first(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    with sync_query(querier, obj, *args) as rows:
        return next(rows)


async def async_query_first(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    async with async_query(querier, obj, *args) as rows:
        return await anext(rows)


@ta.overload
def query_first(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    ...


@ta.overload
def query_first(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[Row]:
    ...


@_sync_async_query_func(sync_query_first, async_query_first)
def query_first(*args, **kwargs):
    raise RuntimeError


##


def sync_query_opt_first(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    with sync_query(querier, obj, *args) as rows:
        return next(rows, None)


async def async_query_opt_first(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    async with async_query(querier, obj, *args) as rows:
        return await anext(rows, None)


@ta.overload
def query_opt_first(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    ...


@ta.overload
def query_opt_first(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[Row | None]:
    ...


@_sync_async_query_func(sync_query_opt_first, async_query_opt_first)
def query_opt_first(*args, **kwargs):
    raise RuntimeError


##


def sync_query_one(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    with sync_query(querier, obj, *args) as rows:
        return check.single(rows)


async def async_query_one(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    async with async_query(querier, obj, *args) as rows:
        return await check.async_single(rows)


@ta.overload
def query_one(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row:
    ...


@ta.overload
def query_one(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[Row]:
    ...


@_sync_async_query_func(sync_query_one, async_query_one)
def query_one(*args, **kwargs):
    raise RuntimeError


##


def sync_query_opt_one(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    with sync_query(querier, obj, *args) as rows:
        return check.opt_single(rows)


async def async_query_opt_one(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    async with async_query(querier, obj, *args) as rows:
        return await check.async_opt_single(rows)


@ta.overload
def query_opt_one(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> Row | None:
    ...


@ta.overload
def query_opt_one(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[Row | None]:
    ...


@_sync_async_query_func(sync_query_opt_one, async_query_opt_one)
def query_opt_one(*args, **kwargs):
    raise RuntimeError


##


def sync_query_scalar(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Any:
    row = sync_query_one(querier, obj, *args)
    return check.single(row.values)


async def async_query_scalar(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Any:
    row = await async_query_one(querier, obj, *args)
    return check.single(row.values)


@ta.overload
def query_scalar(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Any:
    ...


@ta.overload
def query_scalar(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[ta.Any]:
    ...


@_sync_async_query_func(sync_query_scalar, async_query_scalar)
def query_scalar(*args, **kwargs):
    raise RuntimeError


##


def sync_query_maybe_scalar(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> lang.Maybe[ta.Any]:
    row = sync_query_opt_one(querier, obj, *args)
    if row is not None:
        return lang.just(check.single(row.values))
    else:
        return lang.empty()


async def async_query_maybe_scalar(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> lang.Maybe[ta.Any]:
    row = await async_query_opt_one(querier, obj, *args)
    if row is not None:
        return lang.just(check.single(row.values))
    else:
        return lang.empty()


@ta.overload
def query_maybe_scalar(
        querier: Querier,
        obj: ta.Any,
        *args: ta.Any,
) -> lang.Maybe[ta.Any]:
    ...


@ta.overload
def query_maybe_scalar(
        querier: AsyncQuerier,
        obj: ta.Any,
        *args: ta.Any,
) -> ta.Awaitable[lang.Maybe[ta.Any]]:
    ...


@_sync_async_query_func(sync_query_maybe_scalar, async_query_maybe_scalar)
def query_maybe_scalar(*args, **kwargs):
    raise RuntimeError



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import typing as ta

from ...lite.namespaces import NamespaceClass
from ..types import BytesLike
from .types import ByteStreamBuffer
from .types import ByteStreamBufferLike
from .types import ByteStreamBufferView


CanByteStreamBuffer = ta.Union[BytesLike, 'ByteStreamBufferLike']  # ta.TypeAlias


##


class ByteStreamBuffers(NamespaceClass):
    _BYTES_TYPES: ta.ClassVar[ta.Tuple[type, ...]] = (
        bytes,
        bytearray,
    )

    _BYTES_LIKE_TYPES: ta.ClassVar[ta.Tuple[type, ...]] = (
        *_BYTES_TYPES,
        memoryview,
    )

    _CAN_CONVERT_TYPES: ta.ClassVar[ta.Tuple[type, ...]] = (
        *_BYTES_LIKE_TYPES,
        ByteStreamBufferLike,
    )

    #

    @classmethod
    def can_bytes(cls, obj: ta.Any, /) -> bool:
        return type(obj) in (cts := cls._CAN_CONVERT_TYPES) or isinstance(obj, cts)

    #

    @classmethod
    @ta.overload
    def to_bytes(
            cls,
            obj: ta.Any,
            /, *,
            strict: ta.Literal[True],
            or_none: ta.Literal[True],
    ) -> ta.Optional[bytes]:
        ...

    @classmethod
    @ta.overload
    def to_bytes(
            cls,
            obj: ta.Any,
            /, *,
            strict: ta.Literal[True],
            or_none: ta.Literal[False] = False,
    ) -> bytes:
        ...

    @classmethod
    @ta.overload
    def to_bytes(
            cls,
            obj: ta.Any,
            /, *,
            strict: ta.Literal[False] = False,
            or_none: ta.Literal[True],
    ) -> ta.Union[bytes, bytearray, None]:
        ...

    @classmethod
    @ta.overload
    def to_bytes(
            cls,
            obj: ta.Any,
            /, *,
            strict: ta.Literal[False] = False,
            or_none: ta.Literal[False] = False,
    ) -> ta.Union[bytes, bytearray]:
        ...

    @classmethod
    def to_bytes(
            cls,
            obj,
            /, *,
            strict=False,
            or_none=False,
    ):
        """
        Returns a non-shared version of the given object. If a possibly shared memoryview is acceptable, use
        `iter_segments`.
        """

        if strict:
            if (ot := type(obj)) is bytes or isinstance(obj, bytes):
                return obj

            elif ot is bytearray:
                return bytes(obj)

            elif isinstance(obj, memoryview):
                return cls.memoryview_to_bytes_strict(obj)

        else:
            if (ot := type(obj)) is bytes or ot is bytearray or isinstance(obj, cls._BYTES_TYPES):
                return obj

            elif isinstance(obj, memoryview):
                return cls.memoryview_to_bytes(obj)

        if isinstance(obj, ByteStreamBufferView):
            return obj.tobytes()

        elif isinstance(obj, ByteStreamBufferLike):
            return b''.join(bytes(mv) for mv in obj.segments())

        elif or_none:
            return None

        else:
            raise TypeError(obj)

    #

    @classmethod
    @ta.overload
    def bytes_len(cls, obj: ta.Any, or_none: ta.Literal[True], /) -> ta.Optional[int]:
        ...

    @classmethod
    @ta.overload
    def bytes_len(cls, obj: ta.Any, or_none: ta.Literal[False] = False, /) -> int:
        ...

    @classmethod
    def bytes_len(cls, obj, or_none=False):
        if cls.can_bytes(obj):
            return len(obj)

        elif or_none:
            return None

        else:
            raise TypeError(obj)

    #

    @staticmethod
    def iter_segments(obj: ta.Any, /) -> ta.Iterator[memoryview]:
        if (ot := type(obj)) is memoryview:
            yield obj
        elif ot is bytes or ot is bytearray:
            yield memoryview(obj)

        elif isinstance(obj, memoryview):
            yield obj
        elif isinstance(obj, (bytes, bytearray)):
            yield memoryview(obj)

        elif isinstance(obj, ByteStreamBufferLike):
            yield from obj.segments()

        else:
            raise TypeError(obj)

    #

    @staticmethod
    def split(buf: ByteStreamBuffer, sep: bytes, /, *, final: bool = False) -> ta.List[ByteStreamBufferView]:
        out: ta.List[ByteStreamBufferView] = []
        while (i := buf.find(sep)) >= 0:
            out.append(buf.split_to(i + 1))
        if final and len(buf):
            out.append(buf.split_to(len(buf)))
        return out

    #

    @classmethod
    def memoryview_to_bytes(cls, mv: memoryview, /) -> ta.Union[bytes, bytearray]:
        if (((ot := type(obj := mv.obj)) is bytes or ot is bytearray or isinstance(obj, cls._BYTES_TYPES)) and len(mv) == len(obj)):  # type: ignore[arg-type]  # noqa
            return obj  # type: ignore[return-value]

        return mv.tobytes()

    @staticmethod
    def memoryview_to_bytes_strict(mv: memoryview, /) -> bytes:
        if (((ot := type(obj := mv.obj)) is bytes or isinstance(obj, bytes)) and len(mv) == len(obj)):  # type: ignore[arg-type]  # noqa
            return obj  # type: ignore[return-value]

        return mv.tobytes()



"""
Objects:
 + Object keys may be an ECMAScript 5.1 IdentifierName.
 + Objects may have a single trailing comma.
Arrays:
 + Arrays may have a single trailing comma.
Strings:
 + Strings may be single quoted.
 + Strings may span multiple lines by escaping new line characters.
 + Strings may include character escapes.
Numbers:
 + Numbers may be hexadecimal.
 + Numbers may have a leading or trailing decimal point.
 + Numbers may be IEEE 754 positive infinity, negative infinity, and NaN.
 + Numbers may begin with an explicit plus sign.
Comments:
 + Single and multi-line comments are allowed.
White Space:
 + Additional white space characters are allowed.
"""
import typing as ta

from ..json.stream.building import JsonValueBuilder
from ..json.stream.lexing import JsonStreamLexer
from ..json.stream.parsing import JsonStreamParser
from ..json.stream.utils import JsonStreamValueParser
from .literals import parse_number_literal
from .literals import parse_string_literal


##


def make_machinery(
        *,
        include_raw: bool = False,
        yield_object_lists: bool = False,
) -> JsonStreamValueParser.Machinery:
    return JsonStreamValueParser.Machinery(
        JsonStreamLexer(
            include_raw=include_raw,

            allow_extended_space=True,

            allow_comments=True,

            allow_single_quotes=True,
            string_literal_parser=parse_string_literal,

            allow_extended_number_literals=True,
            number_literal_parser=parse_number_literal,

            allow_extended_idents=True,
        ),

        JsonStreamParser(
            allow_trailing_commas=True,

            allow_extended_idents=True,
        ),

        JsonValueBuilder(
            yield_object_lists=yield_object_lists,
        ),
    )


def stream_parse_values(i: ta.Iterable[str], **kwargs: ta.Any) -> ta.Iterator[ta.Any]:
    return JsonStreamValueParser.parse_values(make_machinery(**kwargs), i)


def stream_parse_one_value(i: ta.Iterable[str], **kwargs: ta.Any) -> ta.Any:
    return JsonStreamValueParser.parse_one_value(make_machinery(**kwargs), i)


def stream_parse_exactly_one_value(i: ta.Iterable[str], **kwargs: ta.Any) -> ta.Any:
    return JsonStreamValueParser.parse_exactly_one_value(make_machinery(**kwargs), i)



import dataclasses as dc
import typing as ta

from ..textwrap import TextwrapOpts
from .groups import group_indents
from .lists import ListBuilder
from .parts import Part
from .parts import build_root
from .reflowing import TextwrapReflower
from .reflowing import reflow_block_text


##


DEFAULT_TAB_WIDTH: int = 4


def replace_tabs(s: str, tab_width: int | None = None) -> str:
    if tab_width is None:
        tab_width = DEFAULT_TAB_WIDTH
    return s.replace('\t', ' ' * tab_width)


##


def parse(
        s: str,
        *,
        tab_width: int | None = None,
        allow_improper_list_children: bool | ta.Literal['lists_only'] | None = None,
) -> Part:
    s = replace_tabs(
        s,
        tab_width=tab_width,
    )

    root = build_root(s)

    root = group_indents(root)

    root = ListBuilder(
        allow_improper_children=allow_improper_list_children,
    ).build_lists(root)

    return root


##


DEFAULT_TEXTWRAP_OPTS = TextwrapOpts(
    width=120,
    break_on_hyphens=False,
)


def docwrap(
        s: str,
        *,
        width: int | None = None,
        textwrap: TextwrapOpts | ta.Mapping[str, ta.Any] | None = None,
        allow_improper_list_children: bool | ta.Literal['lists_only'] = False,
) -> Part:
    if isinstance(textwrap, ta.Mapping):
        textwrap = TextwrapOpts(**textwrap)
    elif textwrap is None:
        textwrap = DEFAULT_TEXTWRAP_OPTS
    if width is not None:
        textwrap = dc.replace(textwrap, width=width)

    root = parse(
        s,
        allow_improper_list_children=allow_improper_list_children,
    )

    root = reflow_block_text(
        root,
        TextwrapReflower(opts=textwrap),
    )

    return root



import abc

from ... import lang
from ..params import ParamStyle
from .columns import Column


##


class Adapter(lang.Abstract):
    @property
    @abc.abstractmethod
    def param_style(self) -> ParamStyle | None:
        raise NotImplementedError

    @abc.abstractmethod
    def scan_type(self, c: Column) -> type:
        raise NotImplementedError


##


class HasAdapter(lang.Abstract):
    @property
    @abc.abstractmethod
    def adapter(self) -> Adapter:
        raise NotImplementedError



from ... import dataclasses as dc
from ... import lang


##


@dc.dataclass(frozen=True)
class Dtype(lang.Abstract, lang.Sealed):
    pass


@dc.dataclass(frozen=True)
class Integer(Dtype, lang.Singleton):
    pass


@dc.dataclass(frozen=True)
class String(Dtype, lang.Singleton):
    pass


@dc.dataclass(frozen=True)
class Datetime(Dtype, lang.Singleton):
    pass



import collections.abc
import typing as ta

from .errors import DuplicateKeyError


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


def yield_dict_init(*args: ta.Any, **kwargs: ta.Any) -> ta.Iterator[tuple[ta.Any, ta.Any]]:
    if len(args) > 1:
        raise TypeError

    if args:
        [src] = args

        # Prefer .items() as it's potentially faster.
        if isinstance(src, collections.abc.Mapping):
            # Not yield-from, explicitly enforce 2ple-ness
            for k, v in src.items():
                yield (k, v)

        # Support keys() duck-typed dict init behavior:
        #  https://docs.python.org/3/library/stdtypes.html#dict
        #  https://github.com/python/cpython/blob/95d9dea1c4ed1b1de80074b74301cee0b38d5541/Objects/dictobject.c#L2671
        elif hasattr(src, 'keys'):
            # https://github.com/python/cpython/blob/95d9dea1c4ed1b1de80074b74301cee0b38d5541/Objects/dictobject.c#L2927
            for k in src.keys():  # noqa
                yield (k, src[k])

        else:
            for k, v in src:
                yield (k, v)

    yield from kwargs.items()


def merge_dicts(
        *dcts: ta.Mapping[K, V],
        pair_fn: ta.Callable[[K, V], tuple[K, V] | None] | None = None,
        conflict_fn: ta.Callable[[K, V, V], tuple[K, V] | None] | None = None,
) -> dict[K, V]:
    out: dict[K, V] = {}

    for d in dcts:
        for k_v in d.items():
            if pair_fn is not None and (k_v := pair_fn(*k_v)) is None:  # type: ignore[assignment]
                continue

            k, v = k_v
            if k in out:
                if conflict_fn is None:
                    raise DuplicateKeyError(k)

                if (k_v := conflict_fn(k, out[k], v)) is None:  # type: ignore[assignment]
                    continue
                k, v = k_v

            out[k] = v

    return out


##


class _EmptyMap(ta.Mapping[K, V]):
    def __init_subclass__(cls, **kwargs):
        raise TypeError

    def __new__(cls, *args, **kwargs):
        if args or kwargs:
            raise TypeError
        return _EMPTY_MAP

    def __hash__(self) -> int:
        return hash(_EmptyMap)

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'

    def __init__(self) -> None:
        super().__init__()

    def __getitem__(self, k: K) -> V:
        raise KeyError

    def __len__(self) -> int:
        return 0

    def __iter__(self) -> ta.Iterator[K]:
        return
        yield  # noqa


_EMPTY_MAP = object.__new__(_EmptyMap)


def empty_map() -> ta.Mapping[K, V]:
    return _EMPTY_MAP



import functools
import typing as ta

from .base import AsyncLifecycle
from .base import Lifecycle
from .managed import AsyncLifecycleManaged
from .managed import LifecycleManaged


##


@functools.singledispatch
def unwrap_lifecycle(obj: ta.Any) -> Lifecycle | None:
    return None


@unwrap_lifecycle.register
def _(obj: Lifecycle) -> Lifecycle:
    return obj


@unwrap_lifecycle.register
def _(obj: LifecycleManaged) -> Lifecycle:
    return obj._lifecycle  # noqa


#


@functools.singledispatch
def unwrap_async_lifecycle(obj: ta.Any) -> AsyncLifecycle | None:
    return None


@unwrap_async_lifecycle.register
def _(obj: AsyncLifecycle) -> AsyncLifecycle:
    return obj


@unwrap_async_lifecycle.register
def _(obj: AsyncLifecycleManaged) -> AsyncLifecycle:
    return obj._lifecycle  # noqa


#


def unwrap_any_lifecycle(obj: ta.Any) -> Lifecycle | AsyncLifecycle | None:
    if (lc := unwrap_lifecycle(obj)) is not None:
        return lc

    elif (alc := unwrap_async_lifecycle(obj)) is not None:
        return alc

    else:
        return None



# ruff: noqa: UP006 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ....io.pipelines.core import IoPipelineHandler
from ....io.pipelines.core import IoPipelineHandlerContext
from ....io.pipelines.core import IoPipelineMessages
from ....io.pipelines.flow.types import IoPipelineFlowMessages
from ....io.streams.utils import ByteStreamBuffers
from ....lite.abstract import Abstract
from ..objects import IoPipelineHttpMessageBodyData
from ..objects import IoPipelineHttpMessageEnd
from ..objects import IoPipelineHttpMessageHead
from ..objects import IoPipelineHttpMessageObjects
from .codings import DefaultIoPiplineHttpCompressionCodings
from .codings import IoPiplineHttpCompressorCoding
from .codings import IoPiplineHttpCompressorCodings


##


@dc.dataclass(frozen=True)
class IoPipelineHttpCompressionConfig:
    DEFAULT: ta.ClassVar['IoPipelineHttpCompressionConfig']


IoPipelineHttpCompressionConfig.DEFAULT = IoPipelineHttpCompressionConfig()


#


class IoPipelineHttpObjectCompressor(
    IoPipelineHttpMessageObjects,
    IoPipelineHandler,
    Abstract,
):
    def __init__(
            self,
            codings: ta.Optional[IoPiplineHttpCompressorCodings] = None,
            config: IoPipelineHttpCompressionConfig = IoPipelineHttpCompressionConfig.DEFAULT,
    ) -> None:
        super().__init__()

        self._config = config
        if codings is None:
            codings = DefaultIoPiplineHttpCompressionCodings.COMPRESSOR
        self._codings = codings

        self._compressor: ta.Optional[IoPiplineHttpCompressorCoding] = None

    #

    def _reset(self) -> None:
        self._compressor = None

    #

    def _on_outbound_final_output(self, ctx: IoPipelineHandlerContext, msg: IoPipelineMessages.FinalOutput) -> None:
        if self._compressor is None:
            ctx.feed_out(msg)
            return

        self._reset()

        ctx.feed_out(self._make_aborted('eof before end of message'))
        ctx.feed_out(msg)

    def _on_outbound_head(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageHead) -> None:
        if self._compressor is not None:
            ctx.feed_out(self._make_aborted('unexpected message sequence'))
            return

        enc = msg.headers.lower.get('content-encoding', ())

        for coding_name, coding in self._codings.items():
            if coding_name.lower() in enc:
                self._compressor = coding()
                break

        ctx.feed_out(msg)

    def _on_outbound_body_data(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageBodyData) -> None:
        if (z := self._compressor) is None:
            ctx.feed_out(msg)
            return

        for mv in ByteStreamBuffers.iter_segments(msg.data):
            out = z.compress(mv)
            if out:
                ctx.feed_out(self._make_body_data(out))

    def _on_outbound_flush_output(self, ctx: IoPipelineHandlerContext, msg: IoPipelineFlowMessages.FlushOutput) -> None:
        if (z := self._compressor) is None:
            ctx.feed_out(msg)
            return

        if chunk := z.flush():
            ctx.feed_out(self._make_body_data(chunk))

        ctx.feed_out(msg)

    def _on_outbound_end(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageEnd) -> None:
        if (z := self._compressor) is None:
            ctx.feed_out(msg)
            return

        out = z.finish()
        if out:
            ctx.feed_out(self._make_body_data(out))

        self._reset()
        ctx.feed_out(msg)

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineMessages.FinalOutput):
            self._on_outbound_final_output(ctx, msg)

        elif isinstance(msg, self._head_type):
            self._on_outbound_head(ctx, msg)

        elif isinstance(msg, self._body_data_type):
            self._on_outbound_body_data(ctx, msg)

        elif isinstance(msg, IoPipelineFlowMessages.FlushOutput):
            self._on_outbound_flush_output(ctx, msg)

        elif isinstance(msg, self._end_type):
            self._on_outbound_end(ctx, msg)

        else:
            ctx.feed_out(msg)



from ...lite.abstract import Abstract
from .base import CounterLoggerMetric
from .base import HistogramLoggerMetric
from .base import SecondsLoggerMetricUnit


##


class TimeoutLoggerMetric(CounterLoggerMetric):
    pass


class LatencyLoggerMetric(SecondsLoggerMetricUnit, HistogramLoggerMetric, Abstract):
    pass



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import enum
import ssl
import typing as ta

from ...streams.utils import ByteStreamBuffers
from ..bytes.buffering import InboundBytesBufferingIoPipelineHandler
from ..bytes.buffering import OutboundBytesBufferingIoPipelineHandler
from ..core import IoPipelineHandlerContext
from ..core import IoPipelineMessages
from ..flow.types import IoPipelineFlow
from ..flow.types import IoPipelineFlowMessages


##


class SslIoPipelineHandler(
    InboundBytesBufferingIoPipelineHandler,
    OutboundBytesBufferingIoPipelineHandler,
):
    """
    TLS via ssl.MemoryBIO + SSLObject.

    Model:
      - inbound bytes  : TLS records from transport -> feed into in_bio
      - outbound bytes : TLS records to transport   <- drained from out_bio
      - decrypted plaintext bytes are fed inbound (ctx.feed_in)
      - plaintext to encrypt bytes arrive outbound (outbound())

    TODO:
     - overhaul this thing gawd
     - shutdown timeout
     - context.set_alpn_protocols(['http/1.1'])
     - context.post_handshake_auth = True ?
    """

    @dc.dataclass(frozen=True)
    class Config:
        DEFAULT: ta.ClassVar['SslIoPipelineHandler.Config']

        read_chunk_size: int = 64 * 1024

    Config.DEFAULT = Config()

    def __init__(
            self,
            ssl_ctx: ta.Optional[ssl.SSLContext] = None,
            *,
            server_side: bool,
            server_hostname: ta.Optional[str] = None,
            ssl_session: ta.Optional[ssl.SSLSession] = None,
            config: ta.Optional[Config] = None,
    ) -> None:
        super().__init__()

        if ssl_ctx is None:
            ssl_ctx = ssl.create_default_context()
        self._ssl_ctx = ssl_ctx
        self._server_side = server_side
        self._server_hostname = server_hostname
        self._ssl_session = ssl_session
        if config is None:
            config = SslIoPipelineHandler.Config.DEFAULT
        self._config = config

    #

    def inbound_buffered_bytes(self) -> ta.Optional[int]:
        return 0 if self._state is None else self._in_bio.pending

    def outbound_buffered_bytes(self) -> ta.Optional[int]:
        return self._pending_outbound_bytes + (0 if self._state is None else self._out_bio.pending)

    #

    class State(enum.Enum):
        NEW = 'new'
        HANDSHAKE = 'handshake'
        ESTABLISHED = 'established'
        SHUTTING_DOWN = 'shutting_down'
        CLOSED = 'closed'

    _state: State

    @property
    def state(self) -> ta.Optional[State]:
        try:
            return self._state
        except AttributeError:
            return None

    _in_bio: ssl.MemoryBIO
    _out_bio: ssl.MemoryBIO

    _ssl_obj: ssl.SSLObject

    # Plaintext writes that we couldn't complete yet (usually because handshake isn't established, or write hit
    # WANT_READ and we're waiting for peer).
    _pending_plaintext_out: ta.List[ta.Any]
    _pending_outbound_bytes = 0

    _is_ready_for_input: bool = False
    _read_requested = False

    _inbound_eof_sent = False

    def _ensure_state(self) -> State:
        try:
            return self._state
        except AttributeError:
            pass

        self._in_bio = ssl.MemoryBIO()
        self._out_bio = ssl.MemoryBIO()
        self._ssl_obj = self._ssl_ctx.wrap_bio(
            self._in_bio,
            self._out_bio,
            server_side=self._server_side,
            server_hostname=self._server_hostname,
            session=self._ssl_session,
        )

        self._pending_plaintext_out = []

        self._state = self.State.NEW
        return self._state

    #

    def _fc(self, ctx: IoPipelineHandlerContext) -> ta.Optional[IoPipelineFlow]:
        return ctx.services.find(IoPipelineFlow)

    def _is_auto_read(self, ctx: IoPipelineHandlerContext) -> bool:
        if (flow := ctx.services.find(IoPipelineFlow)) is None:
            return True
        return flow.is_auto_read()

    #

    def _maybe_send_ready_for_input(self, ctx: IoPipelineHandlerContext) -> None:
        fc = self._fc(ctx)
        if fc is None or fc.is_auto_read():
            return

        # REASON 1: Handshake is stalled waiting for peer bytes. We must bridge this gap because the 'App' hasn't asked
        # for data yet.
        if self._state in (self.State.NEW, self.State.HANDSHAKE) and self._is_ready_for_input:
            self._is_ready_for_input = False
            if not IoPipelineFlow.is_auto_read(ctx):
                ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())
            return

        # REASON 2: Downstream asked for data (_read_requested), but our internal BIO and SSL buffers are dry
        # (_is_ready_for_input).
        if self._state == self.State.ESTABLISHED and self._read_requested and self._is_ready_for_input:
            # Only request more from transport if the BIO isn't already EOF-ed
            if self._in_bio.pending == 0 and not self._in_bio.eof:
                self._is_ready_for_input = False
                # We don't reset _read_requested here! We keep it True so that when the bytes eventually arrive,
                # inbound() knows to pump them.
                if not IoPipelineFlow.is_auto_read(ctx):
                    ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())

    def _pump_plaintext_in(self, ctx: IoPipelineHandlerContext) -> bool:
        """Read decrypted data from SSLObject -> inbound plaintext."""

        # We pump if we have a reason to:
        # 1. Auto-read is on.
        # 2. Downstream explicitly asked (_read_requested).
        # 3. We are handshaking (SSL engine is the consumer).

        produced = False

        while self._is_auto_read(ctx) or self._read_requested or self._state != self.State.ESTABLISHED:
            try:
                b = self._ssl_obj.read(self._config.read_chunk_size)

            except ssl.SSLWantReadError:
                # The SSL machine needs more ciphertext from the wire to progress.
                self._is_ready_for_input = True
                break

            except ssl.SSLWantWriteError:
                # Renegotiation or handshake needs to flush TLS records.
                self._pump_tls_out(ctx)
                break

            except ssl.SSLZeroReturnError:
                b = b''

            except ssl.SSLError as e:  # noqa
                # Robustness: fatal SSL errors should propagate to blow up the pipeline.
                raise

            if not b:
                # 1. Clear the manual read token
                if not self._is_auto_read(ctx):
                    self._read_requested = False

                # 2. Emit the synthetic FinalInput EXACTLY ONCE
                if not self._inbound_eof_sent:
                    self._inbound_eof_sent = True
                    ctx.feed_in(IoPipelineMessages.FinalInput())
                    # A FinalInput satisfies the read, but we DO NOT want to trigger a FlushInput in outbound().
                    return False

                break  # Already sent EOF, stop pumping

            # SUCCESS: We got plaintext.
            produced = True
            if self._state == self.State.ESTABLISHED:
                # Satisfy the 'ReadyForInput' token.
                if not self._is_auto_read(ctx):
                    self._read_requested = False

                ctx.feed_in(b)

                # In manual mode, we stop after one successful emission to satisfy Netty's 'one-read-per-request'
                # contract.
                if not self._is_auto_read(ctx):
                    return True

        return produced

    def _pump_tls_out(self, ctx: IoPipelineHandlerContext) -> None:
        """Drain out_bio -> outbound TLS records."""

        fo: ta.List[ta.Any] = []

        n = 0
        while True:
            b = self._out_bio.read()
            if not b:
                break
            fo.append(b)
            n += 1

        if self._out_bio.eof and self._state != self.State.CLOSED:
            # The SSL engine has finished its shutdown sequence (sent close_notify).
            self._state = self.State.CLOSED
            self._pending_plaintext_out.clear()
            self._pending_outbound_bytes = 0

        if (
                self._state in (self.State.HANDSHAKE, self.State.CLOSED) and
                n and
                self._fc(ctx) is not None
        ):
            fo.append(IoPipelineFlowMessages.FlushOutput())

        if self._state == self.State.CLOSED and self._pending_final_output is not None:
            fo.append(self._pending_final_output)
            self._pending_final_output = None

        for msg in fo:
            ctx.feed_out(msg)

    def _handshake_step(self, ctx: IoPipelineHandlerContext) -> bool:
        """Try to advance handshake. Returns True if established. Always pumps TLS output after attempting handshake."""

        if self._state == self.State.CLOSED:
            return False

        if self._state == self.State.NEW:
            self._state = self.State.HANDSHAKE

        if self._state == self.State.HANDSHAKE:
            try:
                self._ssl_obj.do_handshake()

            except ssl.SSLWantReadError:
                # Need peer bytes; but we also may have generated output.
                self._pump_tls_out(ctx)
                self._is_ready_for_input = True
                return False

            except ssl.SSLWantWriteError:
                # Need to flush our pending handshake records.
                self._pump_tls_out(ctx)
                return False

            self._state = self.State.ESTABLISHED
            # Handshake completion can still leave TLS output pending.
            self._pump_tls_out(ctx)
            return True

        return self._state == self.State.ESTABLISHED

    def _write_plaintext(self, ctx: IoPipelineHandlerContext, data: ta.Any) -> bool:
        """
        Write plaintext into SSLObject (encrypt). Returns True if fully accepted.
        On WANT_READ, caller should wait for inbound TLS.
        On WANT_WRITE, caller should flush out_bio.
        """

        for mv in ByteStreamBuffers.iter_segments(data):
            while True:
                try:
                    self._ssl_obj.write(mv)
                    break

                except ssl.SSLWantWriteError:
                    self._pump_tls_out(ctx)
                    # retry same mv
                    continue

                except ssl.SSLWantReadError:
                    self._is_ready_for_input = True
                    return False

        return True

    def _drain_pending_plaintext_out(self, ctx: IoPipelineHandlerContext) -> None:
        if not self._pending_plaintext_out:
            return
        if self._state != self.State.ESTABLISHED:
            return

        # Try to push everything; if we get WANT_READ mid-way, keep remainder.
        new_q: ta.List[ta.Any] = []
        new_b = 0
        for item in self._pending_plaintext_out:
            ok = self._write_plaintext(ctx, item)
            self._pump_tls_out(ctx)
            if not ok:
                new_q.append(item)
                new_b += len(item)
        self._pending_plaintext_out = new_q
        self._pending_outbound_bytes = new_b

    #

    _pending_final_output: ta.Any = None

    def _on_inbound_flush_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineFlowMessages.FlushInput) -> None:  # noqa
        self._ensure_state()

        self._maybe_send_ready_for_input(ctx)

        if self._state in (self.State.NEW, self.State.HANDSHAKE):
            return

        ctx.feed_in(msg)

    def _on_inbound_final_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineMessages.FinalInput) -> None:  # noqa
        self._ensure_state()

        # Signal EOF to the incoming BIO; this is the correct way to tell SSLObject there will be no more peer
        # bytes.
        try:
            self._in_bio.write_eof()
        except Exception as e:  # FIXME:  # noqa
            raise

        # Drain what we can: may emit decrypted tail, and/or TLS close_notify output.
        self._handshake_step(ctx)
        self._pump_plaintext_in(ctx)
        self._pump_tls_out(ctx)

        # If the engine is still technically self.State.ESTABLISHED but the transport is gone, we must force a
        # transition to prevent stuck buffers.
        if self._state != self.State.CLOSED:
            self._state = self.State.CLOSED
            self._pending_plaintext_out.clear()

        # If we reach EOF, any pending manual read request is effectively satisfied by 'nothing'. Clear it so we
        # don't try to ask the already-closed transport for more.
        self._read_requested = False

        # If an abrupt disconnect happened, ensure downstream gets an EOF
        if not self._inbound_eof_sent:
            self._inbound_eof_sent = True
            ctx.feed_in(IoPipelineMessages.FinalInput())

        # CONSUME the transport's FinalInput to prevent duplicate EOFs downstream
        ctx.mark_propagated('inbound', msg)

    def _on_inbound_bytes(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._ensure_state()

        # Feed ciphertext into in_bio.
        for mv in ByteStreamBuffers.iter_segments(msg):
            self._in_bio.write(mv)

        # Advance handshake if needed.
        established = self._handshake_step(ctx)

        if self._state == self.State.SHUTTING_DOWN:
            try:
                self._ssl_obj.unwrap()
                self._state = self.State.CLOSED
            except (ssl.SSLWantReadError, ssl.SSLWantWriteError):
                pass
            except Exception:  # FIXME:  # noqa
                self._state = self.State.CLOSED

        if established:
            # If we had buffered outbound plaintext while handshaking, try it now.
            self._drain_pending_plaintext_out(ctx)
            if self._fc(ctx) is not None:
                ctx.feed_out(IoPipelineFlowMessages.FlushOutput())

        # Always pump: decrypted plaintext inbound, then any TLS output outbound.
        self._pump_plaintext_in(ctx)
        self._pump_tls_out(ctx)

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineFlowMessages.FlushInput):
            self._on_inbound_flush_input(ctx, msg)
            return

        # Underlying transport EOF.
        if isinstance(msg, IoPipelineMessages.FinalInput):
            self._on_inbound_final_input(ctx, msg)
            return

        if ByteStreamBuffers.can_bytes(msg):
            self._on_inbound_bytes(ctx, msg)
            return

        ctx.feed_in(msg)

    #

    def _on_outbound_ready_for_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineFlowMessages.ReadyForInput) -> None:  # noqa
        self._ensure_state()

        # Mark that downstream wants data
        self._read_requested = True

        if self._state == self.State.ESTABLISHED:
            # Check if we actually satisfied the request
            if self._pump_plaintext_in(ctx):
                if not self._is_auto_read(ctx):
                    ctx.feed_in(IoPipelineFlowMessages.FlushInput())
                return  # Swallow

        if self._state == self.State.CLOSED:
            return

        # If we couldn't satisfy it (or we are handshaking), pass it up.
        ctx.feed_out(msg)

    def _on_outbound_flush_output(self, ctx: IoPipelineHandlerContext, msg: IoPipelineFlowMessages.FlushOutput) -> None:  # noqa
        ctx.feed_out(msg)

    def _on_outbound_final_output(self, ctx: IoPipelineHandlerContext, msg: IoPipelineMessages.FinalOutput) -> None:  # noqa
        self._ensure_state()

        if self._state in (self.State.ESTABLISHED, self.State.HANDSHAKE):
            try:
                # This initiates the TLS close_notify alert
                self._ssl_obj.unwrap()
                self._state = self.State.CLOSED

            except (ssl.SSLWantReadError, ssl.SSLWantWriteError) as se:
                self._state = self.State.SHUTTING_DOWN
                if isinstance(se, ssl.SSLWantReadError):
                    self._is_ready_for_input = False
                    if not IoPipelineFlow.is_auto_read(ctx):
                        ctx.feed_out(IoPipelineFlowMessages.ReadyForInput())
                # Save the FinalOutput; we'll feed it once unwrap completes.
                self._pending_final_output = msg
                ctx.mark_propagated('outbound', msg)
                self._pump_tls_out(ctx)
                return

            except Exception:  # FIXME:  # noqa
                pass

            self._pump_tls_out(ctx)

        ctx.feed_out(msg)

    def _on_outbound_bytes(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        self._ensure_state()

        if self._state == self.State.CLOSED:
            # Handshake or close_notify already finalized. We can't encrypt more; drop or raise.
            self._pending_plaintext_out.clear()
            self._pending_outbound_bytes = 0
            return

        established = self._handshake_step(ctx)

        if not established:
            self._pending_plaintext_out.append(msg)
            self._pending_outbound_bytes += len(msg)
            self._maybe_send_ready_for_input(ctx)
            return

        # Write and check for WANT_READ (Renegotiation)
        ok = self._write_plaintext(ctx, msg)
        self._pump_tls_out(ctx)

        if not ok:
            self._pending_plaintext_out.append(msg)
            self._pending_outbound_bytes += len(msg)

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineFlowMessages.ReadyForInput):
            self._on_outbound_ready_for_input(ctx, msg)
            return

        if isinstance(msg, IoPipelineFlowMessages.FlushOutput):
            self._on_outbound_flush_output(ctx, msg)
            return

        if isinstance(msg, IoPipelineMessages.FinalOutput):
            self._on_outbound_final_output(ctx, msg)
            return

        if ByteStreamBuffers.can_bytes(msg):
            self._on_outbound_bytes(ctx, msg)
            return

        ctx.feed_out(msg)



"""
TODO:
 - squash literal unions - typing machinery doesn't
"""
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import reflect as rfl
from ..api.contexts import MarshalContext
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory
from ..api.values import Value


##


@dc.dataclass(frozen=True)
class LiteralMarshaler(Marshaler):
    e: Marshaler
    vs: frozenset

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        return self.e.marshal(ctx, check.in_(o, self.vs))


class LiteralMarshalerFactory(MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, rfl.Literal) and len(set(map(type, rty.args))) == 1):
            return None
        ety = check.single(set(map(type, rty.args)))
        return lambda: LiteralMarshaler(ctx.make_marshaler(ety), frozenset(rty.args))


@dc.dataclass(frozen=True)
class LiteralUnmarshaler(Unmarshaler):
    e: Unmarshaler
    vs: frozenset

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        return check.in_(self.e.unmarshal(ctx, v), self.vs)


class LiteralUnmarshalerFactory(UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, rfl.Literal) and len(set(map(type, rty.args))) == 1):
            return None
        ety = check.single(set(map(type, rty.args)))
        return lambda: LiteralUnmarshaler(ctx.make_unmarshaler(ety), frozenset(rty.args))



"""
https://datatracker.ietf.org/doc/html/rfc5234
"""
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import lang
from .base import Op
from .core import CORE_RULES
from .errors import AbnfGrammarParseError
from .grammars import Channel
from .grammars import Grammar
from .grammars import Rule
from .matches import Match
from .ops import Either
from .ops import Repeat
from .ops import concat
from .ops import either
from .ops import literal
from .ops import option
from .ops import repeat
from .ops import rule
from .opto import optimize_grammar
from .utils import filter_match_channels
from .utils import fix_ws
from .utils import only_match_rules
from .visitors import RuleMatchVisitor


##


META_GRAMMAR_RULES: ta.Sequence[Rule] = [

    Rule(
        'rulelist',
        repeat(
            1,
            either(
                rule('rule'),
                concat(
                    repeat(
                        rule('c-wsp'),
                    ),
                    rule('c-nl'),
                ),
            ),
        ),
    ),

    Rule(
        'rule',
        concat(
            rule('rulename'),
            rule('defined-as'),
            rule('elements'),
            rule('c-nl'),
        ),
    ),

    Rule(
        'rulename',
        concat(
            rule('ALPHA'),
            repeat(
                either(
                    rule('ALPHA'),
                    rule('DIGIT'),
                    literal('-'),
                ),
            ),
        ),
    ),

    Rule(
        'defined-as',
        concat(
            repeat(
                rule('c-wsp'),
            ),
            either(
                literal('=/'),
                literal('='),
            ),
            repeat(
                rule('c-wsp'),
            ),
        ),
    ),

    Rule(
        'elements',
        concat(
            rule('alternation'),
            repeat(
                rule('c-wsp'),
            ),
        ),
    ),

    Rule(
        'c-wsp',
        either(
            rule('WSP'),
            concat(
                rule('c-nl'),
                rule('WSP'),
            ),
        ),
        channel=Channel.SPACE,
    ),

    Rule(
        'c-nl',
        either(
            rule('comment'),
            rule('CRLF'),
        ),
        channel=Channel.SPACE,
    ),

    Rule(
        'comment',
        concat(
            literal(';'),
            repeat(
                either(
                    rule('WSP'),
                    rule('VCHAR'),
                ),
            ),
            rule('CRLF'),
        ),
        channel=Channel.COMMENT,
    ),

    Rule(
        'alternation',
        concat(
            rule('concatenation'),
            repeat(
                concat(
                    repeat(
                        rule('c-wsp'),
                    ),
                    literal('/'),
                    repeat(
                        rule('c-wsp'),
                    ),
                    rule('concatenation'),
                ),
            ),
        ),
    ),

    Rule(
        'concatenation',
        concat(
            rule('repetition'),
            repeat(
                concat(
                    repeat(
                        1,
                        rule('c-wsp'),
                    ),
                    rule('repetition'),
                ),
            ),
        ),
    ),

    Rule(
        'repetition',
        concat(
            option(
                rule('repeat'),
            ),
            rule('element'),
        ),
    ),

    Rule(
        'repeat',
        either(
            concat(
                repeat(
                    rule('DIGIT'),
                ),
                literal('*'),
                repeat(
                    rule('DIGIT'),
                ),
            ),
            repeat(
                1,
                rule('DIGIT'),
            ),
        ),
    ),

    Rule(
        'element',
        either(
            rule('rulename'),
            rule('group'),
            rule('option'),
            rule('char-val'),
            rule('num-val'),
            rule('prose-val'),
        ),
    ),

    Rule(
        'group',
        concat(
            literal('('),
            repeat(
                rule('c-wsp'),
            ),
            rule('alternation'),
            repeat(
                rule('c-wsp'),
            ),
            literal(')'),
        ),
    ),

    Rule(
        'option',
        concat(
            literal('['),
            repeat(
                rule('c-wsp'),
            ),
            rule('alternation'),
            repeat(
                rule('c-wsp'),
            ),
            literal(']'),
        ),
    ),

    Rule(
        'num-val',
        concat(
            literal('%'),
            either(
                rule('bin-val'),
                rule('dec-val'),
                rule('hex-val'),
            ),
        ),
    ),

    Rule(
        'bin-val',
        concat(
            literal('b'),
            concat(
                repeat(
                    1,
                    rule('BIT'),
                ),
                option(
                    either(
                        repeat(
                            1,
                            concat(
                                literal('.'),
                                repeat(
                                    1,
                                    rule('BIT'),
                                ),
                            ),
                        ),
                        concat(
                            literal('-'),
                            repeat(
                                1,
                                rule('BIT'),
                            ),
                        ),
                    ),
                ),
            ),
        ),
    ),

    Rule(
        'dec-val',
        concat(
            literal('d'),
            concat(
                repeat(
                    1,
                    rule('DIGIT'),
                ),
                option(
                    either(
                        repeat(
                            1,
                            concat(
                                literal('.'),
                                repeat(
                                    1,
                                    rule('DIGIT'),
                                ),
                            ),
                        ),
                        concat(
                            literal('-'),
                            repeat(
                                1,
                                rule('DIGIT'),
                            ),
                        ),
                    ),
                ),
            ),
        ),
    ),

    Rule(
        'hex-val',
        concat(
            literal('x'),
            concat(
                repeat(
                    1,
                    rule('HEXDIG'),
                ),
                option(
                    either(
                        repeat(
                            1,
                            concat(
                                literal('.'),
                                repeat(
                                    1,
                                    rule('HEXDIG'),
                                ),
                            ),
                        ),
                        concat(
                            literal('-'),
                            repeat(
                                1,
                                rule('HEXDIG'),
                            ),
                        ),
                    ),
                ),
            ),
        ),
    ),

    Rule(
        'prose-val',
        concat(
            literal('<'),
            repeat(
                either(
                    literal('\x20', '\x3d'),
                    literal('\x3f', '\x7e'),
                ),
            ),
            literal('>'),
        ),
    ),

    # definitions from RFC 7405
    Rule(
        'char-val',
        either(
            rule('case-insensitive-string'),
            rule('case-sensitive-string'),
        ),
    ),

    Rule(
        'case-insensitive-string',
        concat(
            option(
                literal('%i'),
            ),
            rule('quoted-string'),
        ),
    ),

    Rule(
        'case-sensitive-string',
        concat(
            literal('%s'),
            rule('quoted-string'),
        ),
    ),

    Rule(
        'quoted-string',
        concat(
            rule('DQUOTE'),
            repeat(
                either(
                    literal('\x20', '\x21'),
                    literal('\x23', '\x7e'),
                ),
            ),
            rule('DQUOTE'),
        ),
    ),

]


RAW_META_GRAMMAR = Grammar(
    *CORE_RULES,
    *META_GRAMMAR_RULES,
    root='rulelist',
)

META_GRAMMAR = optimize_grammar(
    RAW_META_GRAMMAR,
    inline_channels=(
        Channel.CONTENT,
        Channel.COMMENT,
        Channel.SPACE,
    ),
)


##


class MetaGrammarRuleMatchVisitor(RuleMatchVisitor[ta.Any]):
    def __init__(self, source: str) -> None:
        super().__init__()

        self._source = source

    @dc.dataclass(frozen=True)
    class RuleName(lang.Final):
        s: str

    @dc.dataclass(frozen=True)
    class QuotedString(lang.Final):
        s: str

    class AltRule(ta.NamedTuple):
        rule: Rule

    @RuleMatchVisitor.register('rule')
    def visit_rule_rule(self, m: Match) -> ta.Any:
        rn_m, da_m, el_m = m.children
        da_s = self._source[da_m.start:da_m.end].strip()
        rn = check.isinstance(self.visit_match(rn_m), self.RuleName).s
        el = self.visit_match(el_m)
        r = Rule(rn, el)
        if da_s == '=':
            return r
        elif da_s == '=/':
            return self.AltRule(r)
        else:
            raise ValueError(da_s)

    @RuleMatchVisitor.register('rulename')
    def visit_rulename_rule(self, m: Match) -> ta.Any:
        return self.RuleName(self._source[m.start:m.end])

    @RuleMatchVisitor.register('elements')
    def visit_elements_rule(self, m: Match) -> ta.Any:
        return self.visit_match(check.single(m.children))

    @RuleMatchVisitor.register('alternation')
    def visit_alternation_rule(self, m: Match) -> ta.Any:
        if len(m.children) == 1:
            return self.visit_match(m.children[0])
        else:
            return either(*map(self.visit_match, m.children))

    @RuleMatchVisitor.register('concatenation')
    def visit_concatenation_rule(self, m: Match) -> ta.Any:
        if len(m.children) == 1:
            return self.visit_match(m.children[0])
        else:
            return concat(*map(self.visit_match, m.children))

    @RuleMatchVisitor.register('repetition')
    def visit_repetition_rule(self, m: Match) -> ta.Any:
        if len(m.children) == 2:
            ti_m, el_m = m.children
            ti = check.isinstance(self.visit_match(ti_m), Repeat.Times)
            el = self.visit_match(el_m)
            return repeat(ti, el)
        elif len(m.children) == 1:
            return self.visit_match(m.children[0])
        else:
            raise ValueError(m)

    @RuleMatchVisitor.register('repeat')
    def visit_repeat_rule(self, m: Match) -> ta.Any:
        s = check.non_empty_str(self._source[m.start:m.end])
        if s == '*':
            return Repeat.Times(0)
        elif '*' in s:
            check.state(s.count('*') == 1)
            if s.endswith('*'):
                return Repeat.Times(int(s[:-1]))
            else:
                mi, mx = s.split('*')
                return Repeat.Times(int(mi), int(mx))
        else:
            return Repeat.Times(n := int(s), n)

    @RuleMatchVisitor.register('element')
    def visit_element_rule(self, m: Match) -> ta.Any:
        c = self.visit_match(check.single(m.children))
        if isinstance(c, Op):
            return c
        elif isinstance(c, self.RuleName):
            return rule(c.s)
        else:
            raise TypeError(c)

    @RuleMatchVisitor.register('group')
    def visit_group_rule(self, m: Match) -> ta.Any:
        return self.visit_match(check.single(m.children))

    @RuleMatchVisitor.register('option')
    def visit_option_rule(self, m: Match) -> ta.Any:
        c = self.visit_match(check.single(m.children))
        return option(check.isinstance(c, Op))

    @RuleMatchVisitor.register('num-val')
    def visit_num_val_rule(self, m: Match) -> ta.Any:
        return self.visit_match(check.single(m.children))

    def _parse_num_val(self, s: str, base: int) -> Op:
        if '-' in s:
            check.not_in('.', s)
            lo, hi = [chr(int(p, base)) for p in s.split('-')]
            return literal(lo, hi)
        elif '.' in s:
            check.not_in('-', s)
            cs = [chr(int(p, base)) for p in s.split('.')]
            return concat(*[literal(c, c) for c in cs])
        else:
            c = chr(int(s, base))
            return literal(c, c)

    @RuleMatchVisitor.register('dec-val')
    def visit_dec_val_rule(self, m: Match) -> ta.Any:
        return self._parse_num_val(self._source[m.start + 1:m.end], 10)

    @RuleMatchVisitor.register('hex-val')
    def visit_hex_val_rule(self, m: Match) -> ta.Any:
        return self._parse_num_val(self._source[m.start + 1:m.end], 16)

    @RuleMatchVisitor.register('char-val')
    def visit_char_val_rule(self, m: Match) -> ta.Any:
        return self.visit_match(check.single(m.children))

    @RuleMatchVisitor.register('case-sensitive-string')
    def visit_case_sensitive_string_rule(self, m: Match) -> ta.Any:
        c = self.visit_match(check.single(m.children))
        return literal(check.isinstance(c, self.QuotedString).s, case_sensitive=True)

    @RuleMatchVisitor.register('case-insensitive-string')
    def visit_case_insensitive_string_rule(self, m: Match) -> ta.Any:
        c = self.visit_match(check.single(m.children))
        return literal(check.isinstance(c, self.QuotedString).s, case_sensitive=False)

    @RuleMatchVisitor.register('quoted-string')
    def visit_quoted_string_rule(self, m: Match) -> ta.Any:
        check.state(m.end - m.start > 2)
        check.state(self._source[m.start] == '"')
        check.state(self._source[m.end - 1] == '"')
        return self.QuotedString(self._source[m.start + 1:m.end - 1])


##


def parse_grammar(
        source: str,
        *,
        root: str | None = None,
        no_core_rules: bool = False,
        no_optimize: bool = False,
        **kwargs: ta.Any,
) -> Grammar:
    source = fix_ws(source)

    if (mg_m := META_GRAMMAR.parse(
            source,
            complete=True,
            **kwargs,
    )) is None:
        raise AbnfGrammarParseError(source)

    mg_m = only_match_rules(mg_m)

    mg_m = filter_match_channels(
        mg_m,
        META_GRAMMAR,
        keep=(Channel.STRUCTURE,),
        keep_children=True,
    )

    check.isinstance(mg_m.op, Repeat)

    mg_rmv = MetaGrammarRuleMatchVisitor(source)
    lst = [
        mg_rmv.visit_match(gg_cm)
        for gg_cm in mg_m.children
    ]

    rules: dict[str, Rule] = {}
    for cur in lst:
        if isinstance(cur, Rule):
            check.not_in(cur._name_f, rules)  # noqa
            rules[cur._name_f] = cur  # noqa
        elif isinstance(cur, MetaGrammarRuleMatchVisitor.AltRule):
            xr = rules[cur.rule._name_f]  # noqa
            xo = xr.op
            if isinstance(xo, Either):
                xo = Either(*xo.children, cur.rule.op, first_match=xo.first_match)
            else:
                xo = Either(xo, cur.rule.op)
            xr = Rule(
                xr._name_f,  # noqa
                xo,
                channel=xr.channel,
            )
            rules[xr._name_f] = xr  # noqa
        else:
            raise TypeError(cur)

    if not no_core_rules:
        for cr in CORE_RULES:
            if cr._name_f not in rules:  # noqa
                rules[cr._name_f] = cr  # noqa

    gram = Grammar(
        *rules.values(),
        root=root,
    )

    if not no_optimize:
        gram = optimize_grammar(gram)

    return gram



# @omlish-lite
# ruff: noqa: UP006 UP007 UP045
import dataclasses as dc
import os.path
import tempfile
import typing as ta

from ..lite.cached import cached_nullary
from ..subprocesses.run import SubprocessRun
from ..subprocesses.run import SubprocessRunnable
from ..subprocesses.run import SubprocessRunOutput
from .ssl import SslCert


##


class TempSslCert(ta.NamedTuple):
    cert: SslCert
    temp_dir: str


@dc.dataclass(frozen=True)
class TempSslCertGenerator(SubprocessRunnable[TempSslCert]):
    @cached_nullary
    def temp_dir(self) -> str:
        return tempfile.mkdtemp()

    @cached_nullary
    def make_run(self) -> SubprocessRun:
        return SubprocessRun.of(
            'openssl',
            'req',
            '-x509',
            '-newkey', 'rsa:2048',

            '-keyout', 'key.pem',
            '-out', 'cert.pem',

            '-days', '365',

            '-nodes',

            '-subj', '/CN=localhost',
            '-addext', 'subjectAltName = DNS:localhost,IP:127.0.0.1',

            cwd=self.temp_dir(),
            capture_output=True,
            check=False,
        )

    def handle_run_output(self, proc: SubprocessRunOutput) -> TempSslCert:
        if proc.returncode:
            raise RuntimeError(f'Failed to generate temp ssl cert: {proc.stderr=}')

        key_file = os.path.join(self.temp_dir(), 'key.pem')
        cert_file = os.path.join(self.temp_dir(), 'cert.pem')
        for file in [key_file, cert_file]:
            if not os.path.isfile(file):
                raise RuntimeError(f'Failed to generate temp ssl cert (file not found): {file}')

        return TempSslCert(
            SslCert(
                key_file=key_file,
                cert_file=cert_file,
            ),
            temp_dir=self.temp_dir(),
        )


def generate_temp_localhost_ssl_cert() -> TempSslCert:
    return TempSslCertGenerator().run()



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import typing as ta

from ....lite.abstract import Abstract
from ..core import IoPipelineHandlerRef


##


class IoPipelineScheduling(Abstract):
    class Handle(Abstract):
        @abc.abstractmethod
        def cancel(self) -> None:
            raise NotImplementedError

    @abc.abstractmethod
    def schedule(
            self,
            handler_ref: IoPipelineHandlerRef,
            delay_s: float,
            fn: ta.Callable[[], None],
    ) -> Handle:
        raise NotImplementedError

    @abc.abstractmethod
    def cancel_all(self, handler_ref: ta.Optional[IoPipelineHandlerRef] = None) -> None:
        raise NotImplementedError
